"""Functions for constructing and analyzing causal graphs."""
from collections.abc import Iterable
import types

import dataclasses

from autograd.extend import primitive
from autograd.numpy.numpy_boxes import ArrayBox
from autograd.tracer import Box, isbox, new_box, Node, trace_stack
import networkx as nx
import numpy as np

import whynot as wn

# Allow tracing to use ints and bools (since we only care about the forward
# pass and not derivatives).
for type_ in [bool, np.bool, np.bool_, int, np.int32, np.int64]:
    ArrayBox.register(type_)


class FunctionBox(Box):
    """Generic box class to encapsulate functions and record their invocation."""

    @primitive
    def __call__(func, *args, **kwargs):
        """Execute function and record its invocation in the graph."""
        return func(*args, **kwargs)


FunctionBox.register(types.FunctionType)


class TracerNode(Node):
    """Base case for constructing causal graphs via computation graphs.

    A TracerNode is a node in the computation graph and symbolically
    represents the application of a function to its parent nodes to
    produce a child node.

    This class is used internally by autograd, and we likely do not every
    need to call it explicitly.
    """

    def __init__(self, value, fun, args, kwargs, parent_argnums, parents):
        """Construct a node out of a function application."""
        # pylint:disable-msg=super-init-not-called
        self.parents = parents
        self.value = value
        self.fun = fun
        self.args = args
        self.kwargs = kwargs
        self.parent_argnums = parent_argnums
        # Only root names are named
        self.name = None

    def initialize_root(self, arg, name=None):
        # pylint:disable-msg=arguments-differ
        """Construct the root of the computation graph.

        Parameters
        ----------
            arg: the initial value
            name: (Optional) the name of the root node.

        """
        self.parents = []
        self.value = arg
        self.fun = None
        self.args = []
        self.kwargs = {}
        self.parent_argnums = None
        self.name = name


def dataclass_to_box(dataclass, trace, name_suffix=None, skip_args=None):
    """Convert a dataclass object to a traceable box.

    Parameters
    ----------
        dataclass: instance of dataclasses.dataclass
            Either a whynot State or a Config object
        trace: trace_stack context manager
            Every box is created within the context of a trace
        name_suffix: str
            (Optional) Suffix to append to every field name of the dataclass
        skip_args: list
            (Optional) List of arguments/dataclass field names to skip tracing.

    Returns
    -------
        box: instance of dataclasses.dataclass
            The same input object with fields replaced by boxes to allow for
            dependency tracing.
        nodes: dict
            Dictionary mapping a root TracerNode to the corresponding Box.
            The name of the TracerNode is the fieldname and the optional suffix.

    """
    # Wrap each input argument in a "box" with an associated start node to allow tracing
    flattened = dataclasses.astuple(dataclass)
    names = [field.name for field in dataclasses.fields(dataclass)]

    suffix = f"_{name_suffix}" if name_suffix else ""
    replacements, node_map = {}, {}
    for name, value in zip(names, flattened):
        if skip_args and name in skip_args:
            # Don't wrap this argument in a box
            replacements[name] = value
        elif hasattr(value, "get_boxable"):
            # The value itself isn't traceble, but it exports
            # methods that allow it to be traced.
            boxable = value.get_boxable()
            boxable = check_and_cast_args(boxable)

            node = TracerNode.new_root(boxable, name + suffix)
            box = new_box(boxable, trace, node)

            value.set_boxable(box)

            replacements[name] = value
            node_map[node] = box

        else:
            # Try to wrap the value in a box, casting as possible for dependency tracing
            value = check_and_cast_args(value)
            node = TracerNode.new_root(value, name + suffix)
            box = new_box(value, trace, node)

            replacements[name] = box
            node_map[node] = box

    # Generate a new dataclass, replacing each value with the corresponding box
    dataclass = dataclasses.replace(dataclass, **replacements)

    return dataclass, node_map


def run_to_box(run, trace_ctx, skip_args=None):
    """Convert a rollout of a dynamical system simulator to a collection of traceable elements.

    Parameters
    ----------
        run: whynot.dynamics.Run
            The run for the user to trace.
        trace_ctx:  trace_stack context manager
            Every box is created within the context of this trace
        skip_args: list
            (Optional) List of fields in the state to avoid tracing.
            These fields are not wrapped in box objects.

    Returns
    -------
        boxed_run: whynot.dynamics.Run
            Run object where each State has all of its fields (except those
            listed in skip_args) replaced with a Box object for tracing.

        global_nodemap: dict
            Dictionary mapping root TracerNode to the corresponding box
            objects. The name of each TracerNode is statename_time.
            During dependency tracing, if we encounter a node in this
            map, then we have found an "ancestor" of the output node
            as a base element of the run.

    """
    boxed_states, global_nodemap = [], {}
    for state, time in zip(run.states, run.times):
        boxed_state, nodemap = dataclass_to_box(
            state, trace_ctx, name_suffix=str(time), skip_args=skip_args
        )
        boxed_states.append(boxed_state)
        global_nodemap.update(nodemap)
    boxed_run = wn.dynamics.Run(states=boxed_states, times=run.times)
    return boxed_run, global_nodemap


def check_and_cast_args(args):
    """Check and cast arguments for the dependency tracer.

    Since autograd is expecting to compute derivatives, ints are not
    allowable as arguments. However, for our purposes, casting ints
    to floats will not affect dependency tracing.
    """
    # Currently this function only casts int into floats and raises
    # if the element is a string. This can be expanded to be more robust.
    def check_and_cast(arg):
        """Attempt to cast the arg to something traceable. Otherwise, error."""
        if isinstance(arg, str):
            raise ValueError(f"Cannot build causal graph for arg {arg} of type str")
        if isinstance(arg, int):
            arg = float(arg)
        return arg

    if isinstance(args, Iterable):
        return [check_and_cast(arg) for arg in args]
    return check_and_cast(args)


def is_dead_node(node):
    """Check if this node adds a spurious dependency, e.g. 0 * constant."""
    if hasattr(node.fun, "fun") and node.fun.fun == np.multiply:
        # If we multiplied a parent by a constant 0, skip.
        for idx, arg in enumerate(node.args):
            if idx not in node.parent_argnums and arg == 0.0:
                return True
    return False


def is_array_node(node):
    """Check if this node creates a array."""
    if hasattr(node.fun, "fun"):
        return node.fun.fun.__name__ == wn.traceable_numpy.array_from_args.__name__
    return False


def is_getitem_node(node):
    """Check if the node extracts an item from an array."""
    if hasattr(node.fun, "fun"):
        return node.fun.fun.__name__ == "__getitem__"
    return False


def backtrack(output_boxes, input_node_map):
    """Trace the computation graph from output boxes to ancestors in the input.

    Given the output of a function called on boxed inputs, backtrack finds
    which, if any, of the nodes in the input_node_map each output element
    depends on.

    Parameters
    ----------
        output_boxes: list
            List of outputs from a traced function.
        input_node_map: dict
            Dictionary of TracerNode -> box elements representing the potential root
            nodes of the computation graph.

    Returns
    -------
        output_dependencies: dict
            Dictionary mapping output_idx to a (deduplicated) list of nodes
            in the input that output idx depends on. For instance, if we have

            def f(x, y, z):
                return x + y, y + z

            Then, we get output_dependencies {0: [Node_x, Node_y], 1: [Node_y, Node_z]}.

    """
    if not isinstance(output_boxes, Iterable):
        output_boxes = [output_boxes]

    # For each output, figure out which (if any) of the inputs it depends on
    # by solving a graph search problem on the computation graph.
    output_dependencies = dict(
        (output_idx, set()) for output_idx in range(len(output_boxes))
    )
    for idx, output_box in enumerate(output_boxes):
        # If the output isn't a box, then it's independent of all the inputs.
        if isbox(output_box):
            # Use breadth-first search to find parents
            # pylint: disable-msg=protected-access
            queue = [output_box._node]
            while queue:
                node = queue.pop(0)

                # Check if ancestor is an input node
                if node in input_node_map:
                    input_box = input_node_map[node]
                    if output_box._trace == input_box._trace:
                        output_dependencies[idx].add(node)

                # If the node corresponds to an irrelevant dependency,
                # e.g. 0 * constant, we skip it.
                if is_dead_node(node):
                    continue

                # Add the parents to the queue
                for parent in node.parents:
                    # This is a hack to handle a common pattern where the
                    # user returns z = np.array([x1, x2]) from a function and
                    # then we inspect one item, e.g. z[0]. By default, both x1 and x2
                    # are parents of z, but in this case, we can detect that
                    # z[0] only depends on x1. In general, however, if after
                    # constructing z, we apply another function to the array,
                    # e.g. f(z) we cannot detect the subset of x that the result depends on.
                    # This is one of the limitations of using boxes based on autograd.
                    # Note this limitation means that we might add extra
                    # edges to the causal graph. It will never mean we miss an
                    # edge, so the returned graphs are still valid.
                    if is_getitem_node(node) and is_array_node(parent):
                        index = node.args[1]
                        queue.append(parent.parents[index])
                    else:
                        queue.append(parent)

    return dict((k, list(v)) for k, v in output_dependencies.items())


def trace_dependencies(func, args):
    """Trace dependencies of the outputs of a function.

    Parameters
    ----------
        func: function
            The function to trace
        args: List
            List of numerical arguments to trace dependencies for.

    Returns
    -------
        output_dependencies: map
            Dictionary mapping from output_idx to the set of input_idx that the
            output depends on.

    Example
    -------
    .. code-block:: python

        >>> def func(x, y, z):
        ...     return x + y, y * z

        >>> trace_dependencies(func, [4, 2, 1])
        {0: [0, 1], 1: {1, 2}

    """
    if not isinstance(args, Iterable):
        args = [args]

    args = check_and_cast_args(args)

    with trace_stack.new_trace() as trace:

        # Wrap each input argument in a "box" with an associated start node to allow tracing
        input_nodes = [
            TracerNode.new_root(arg, name=idx) for idx, arg in enumerate(args)
        ]
        input_boxes = [
            new_box(arg, trace, start_node)
            for arg, start_node in zip(args, input_nodes)
        ]

        # Execute the function to build the computation graph
        output_boxes = func(*input_boxes)

        input_node_map = dict(zip(input_nodes, input_boxes))
        raw_dependencies = backtrack(output_boxes, input_node_map)

        # convert to output-> [list input_idxs]
        dependencies = {}
        for output_idx, input_nodes in raw_dependencies.items():
            dependencies[output_idx] = [node.name for node in input_nodes]
        return dependencies


def trace_dynamics(dynamics):
    """Generate a function to compute the dependency graph of the dynamics.

    Given the dynamics map for a dynamical system simulator, we construct
    another function that, given the state, time, and config, computes
    how the output states depend on the input states and the configuration
    parameters. We cannot statically trace the dynamics because, in general,
    they may depend on time or the value of the variables in the state or
    config.

    Parameters
    ----------
        dynamics: function
            A dynamics map that has the signature def dynamics(state, time,
            config), where state is a tuple, and config is a dataclass. The
            function should return the derivative of the state as an iterable.

    Returns
    -------
        dynamics_tracer: function
            A function which, given an instantiated state, time, and config
            for the simulator, computes the dependency graph as a dictionary
                {output_idx:
                        {"states": [input_state_dependencies],
                        "configs": [config_dependencies]}}

    Example
    -------
    .. code-block:: python

        >>> def dynamics(state, time, config):
        ...    x_1, x_2 = state
        ...    if time < 10:
        ...        return x1, config.param
        ...    return x1 + x2, x2

        >>> tracer = dynamics_tracer(dynamics)
        >>> tracer(state, 0, config)
        {0: {"states": ["x1"], config: []},
         1: {"states": [], config: ["param"]},
        >>> tracer(state, 20, config)
        {0: {"states": ["x1", "x2"], config: []},
         1: {"states": ["x2"], config: []},

    """

    def tracer(state, time, config):
        """Compute output dependencies of the dynamics."""
        with trace_stack.new_trace() as trace:
            state_box, state_node_map = wn.causal_graphs.dataclass_to_box(state, trace)
            config_box, config_node_map = wn.causal_graphs.dataclass_to_box(
                config, trace, skip_args=["start_time", "end_time", "delta_t"]
            )

            # Aggregrate the state and config maps
            node_map = state_node_map.copy()
            node_map.update(config_node_map)

            # Run the dynamics forward to construct the computation graph
            outbox = dynamics(state=state_box.values(), time=time, config=config_box)

            # Trace the output dependencies to nodes in the input
            dependencies = wn.causal_graphs.backtrack(outbox, node_map)

            # Generate a map from output name to input and config dependency names
            names = [f.name for f in dataclasses.fields(state)]
            named_dependencies = {}
            for output_idx, input_nodes in dependencies.items():
                state_deps, config_deps = [], []
                for node in input_nodes:
                    if node in state_node_map:
                        state_deps.append(node.name)
                    else:
                        config_deps.append(node.name)
                named_dependencies[names[output_idx]] = {
                    "states": state_deps,
                    "configs": config_deps,
                }

            return named_dependencies

    return tracer


def build_dynamics_graph(simulator, runs, config, config_nodes=False):
    """Build a graph of the dynamics for a collection of runs.

    This feature is still experiment is currently only supported on a
    handful of simulators, namely HIV, lotka_volterra, and opioid.

    Parameters
    ----------
        simulator: whynot simulator module
            Simulator dynamics to trace, e.g. whynot.hiv.
        runs: list
            List of whynot.dynamics.Runs objects generated by the simulator
        config: whynot.dynamics.BaseConfig
            Config object use for all of the simulator runs
        config_nodes: bool
            Whether or not to add nodes and edges to the graphs for
            configuration variables or to leave the explicit.

    Returns
    -------
        graph: networkx.DiGraph
            Causal graph representing the dynamics for the collection of runs.

    Example
    -------
    .. code-block:: python

        >>> import whynot as wn
        >>> config = wn.hiv.Config(delta=1., end_time=1.)
        >>> runs = [wn.hiv.simulate(wn.hiv.State(),  config)]
        >>> graph = wn.causal_graphs.build_dynamics_graph(wn.hiv, runs, config)
        >>> print(list(graph.nodes))
        uninfected_T1_0.0, infected_T1_0.0, ..., immune_response1.0

    """
    if not simulator.SUPPORTS_CAUSAL_GRAPHS:
        raise ValueError(
            "Simulator does not currently support causal graph construction."
        )

    dynamics_tracer = trace_dynamics(simulator.dynamics)

    # TODO: Eventually we should take the union over all of the discovered
    # edges.  For now, just take the edges from the first run.
    run = runs[0]

    graph = nx.DiGraph()

    state_names = run.initial_state.variable_names()
    config_name = config.parameter_names()

    # Add a node for every state/parameter at every timestep
    for time in run.times:
        graph.add_nodes_from([f"{name}_{time}" for name in state_names])
        if config_nodes:
            graph.add_nodes_from([f"PARAM:{name}_{time}" for name in config_name])

    # Add edges by tracing the dynamics. Assumes the connectivity pattern
    # is constant between each time step (but can generally vary with time).
    for idx, start_state in enumerate(run.states[:-1]):
        start_time, end_time = run.times[idx], run.times[idx + 1]
        dependency_map = dynamics_tracer(start_state, start_time, config)
        for output, dependencies in dependency_map.items():
            output_name = f"{output}_{end_time}"
            graph.add_edges_from(
                [
                    (f"{state}_{start_time}", output_name)
                    for state in dependencies["states"]
                ]
            )
            if config_nodes:
                graph.add_edges_from(
                    [
                        (f"PARAM:{conf_name}_{start_time}", output_name)
                        for conf_name in dependencies["configs"]
                    ]
                )

    return graph


def ate_graph_builder(
    simulator,
    run,
    config,
    intervention,
    treatment_dependencies,
    covariate_dependencies,
    outcome_dependencies,
):
    """Build a causal graph for an average treatment effect estimation experiment.

    Parameters
    ----------
        simulator: whynot.simulator module
            The simulator used to conduct the experiment.
        run: whynot.dynamics.Run
            A single run object generated from the simulator under the config
            and intervention.
        config: whynot.dynamics.BaseConfig
            Configuration object used for the experiment.
        intervention: whynot.dynamics.BaseIntervention
            Intervention object used for the experiment.
        treatment_dependencies: list
            List of node names treatment depends on.
        covariate_dependencies: dict
            Dictionary from covariate_idx to a list of nodes the covariate
            depends on.
        outcome_dependencies: list
            List of node names the outcome variable depends on.

    Returns
    -------
        ate_graph: networkx.DiGraph
            Directed graph corresponding to the causal inference experiment.
            The nodes in the graph include:
                1) State variables from the unrolled dynamics
                2) Configuration parameters at each time step
                3) The treatment node, and
                4) The outcome node.
            Edges in the graph include:
                1) Edges between state nodes from the dynamics
                2) Edges between config and state nodes if a state variable
                   depends on a parameter
                3) Edges between state nodes and the treatment
                4) Edges between the treatment and all of the parameters/nodes
                   that are intervened upon.
                5) Edges between the state nodes and the outcome.

            Finally, the graph has an attribute graph["covariate_names"]
            that record which covariate corresponds to which state nodes.

    """
    graph = build_dynamics_graph(simulator, [run], config, config_nodes=True)

    # All nodes are unobserved by default
    for node in graph.nodes:
        graph.nodes[node]["observed"] = "no"

    # Treatment nodes
    graph.add_node("Treatment", observed="yes")
    # Incoming edges
    graph.add_edges_from([(dep, "Treatment") for dep in treatment_dependencies])
    # Outgoing edges
    intervention_nodes = []
    for time in run.times:
        if time >= intervention.time:
            intervention_nodes.extend(
                [f"PARAM:{param}_{time}" for param in intervention.updates]
            )
    graph.add_edges_from([("Treatment", node) for node in intervention_nodes])

    # Outcome nodes
    graph.add_node("Outcome", observed="yes")
    graph.add_edges_from([(node, "Outcome") for node in outcome_dependencies])

    covariate_names = ["" for _ in range(len(covariate_dependencies))]
    for idx, dependencies in covariate_dependencies.items():
        if len(dependencies) > 1:
            error_msg = (
                "Each covariate should depend on a single node, but covariate "
                f"{idx} depends on {len(dependencies)} nodes!"
            )
            raise ValueError(error_msg)
        name = dependencies[0]
        graph.nodes[name]["observed"] = "yes"
        covariate_names[idx] = name

    graph.graph["covariate_names"] = covariate_names

    return graph



"""Experiments for lending simulator."""

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import delayed_impact
from whynot.simulators.delayed_impact.simulator import INV_CDFS, GROUP_SIZE_RATIO

__all__ = ["get_experiments", "CreditBureauExperiment"]


def get_experiments():
    """Return all experiments for delayed impact simulator."""
    return [CreditBureauExperiment]


################################
# Credit Intervention Experiment
################################
def sample_initial_states(rng):
    """Sample initial states according to FICO data on group membership."""
    group = int(rng.uniform() < GROUP_SIZE_RATIO[1])
    # Compute score via inverse CDF trick
    score = INV_CDFS[group](rng.uniform())
    return delayed_impact.State(group=group, credit_score=score)


def creditscore_threshold(score):
    """Alternate credit bureau scoring policy."""
    return max(score, 600)


def extract_outcomes(run):
    """Outcome is both the score change Delta after 1 step."""
    return run.states[1].credit_score - run.states[0].credit_score


@parameter(
    name="threshold_g0", default=650, description="Lending threshold for group 0",
)
@parameter(
    name="threshold_g1", default=650, description="Lending threshold for group 1",
)
def construct_config(threshold_g0, threshold_g1):
    """Experimental config is parameterized by the lending thresholds."""
    return delayed_impact.Config(
        start_time=0, end_time=1, threshold_g0=threshold_g0, threshold_g1=threshold_g1
    )


#: Effects of interventions on credit scoring on group score changes
CreditBureauExperiment = DynamicsExperiment(
    name="CreditBureauExperiment",
    description="Intervention on the credit scoring mechanism.",
    simulator=delayed_impact,
    # Run for a single time step
    simulator_config=construct_config,
    # Change the threshold on the first step.
    intervention=delayed_impact.Intervention(
        credit_scorer=creditscore_threshold, time=0
    ),
    state_sampler=sample_initial_states,
    # All units are treated
    propensity_scorer=1.0,
    outcome_extractor=extract_outcomes,
    # Only covariate is group membership
    covariate_builder=lambda run: run.initial_state.group,
)



"""Test causal graph building functionality."""
import copy
import dataclasses
import itertools
import math

import numpy as np
import pytest

import whynot as wn
from whynot.dynamics import BaseConfig, BaseIntervention, BaseState

#################################
# Basic dependency tracing tests.
##################################
def test_dependency_tracing_basic():
    """Basic tests for tracing dependencies."""

    def func(x1, x2):
        delta_x1 = 45 * x1 + 18 * x2
        delta_x2 = 2 * x2
        return [delta_x1, delta_x2, 23]

    dependencies = wn.causal_graphs.trace_dependencies(func, (1.0, 0.2))
    assert set(dependencies[0]) == set([0, 1])
    assert set(dependencies[1]) == set([1])
    assert set(dependencies[2]) == set()

    # Ensure things work for univariate output
    def univar(x1, x2):
        return x1 + x2

    dependencies = wn.causal_graphs.trace_dependencies(univar, (1.0, 0.2))
    assert set(dependencies[0]) == set([0, 1])

    # univariate input
    def univar_input(x1):
        return 2 * x1 ** 2

    dependencies = wn.causal_graphs.trace_dependencies(univar_input, 4)
    assert set(dependencies[0]) == set([0])


def test_noop():
    """Test is the causal graph builder correctly handles noops."""

    def noop(x1):
        return 0 * x1

    dependencies = wn.causal_graphs.trace_dependencies(noop, 1.0)
    assert set(dependencies[0]) == set()

    def noop1(x, y, z):
        a = 2 * x + y - z
        b = x * y * z * 0.0
        c = 0 * a ** 2 - b
        return c

    dependencies = wn.causal_graphs.trace_dependencies(noop1, (1, 2, 3))
    assert set(dependencies[0]) == set()

    dependencies = wn.causal_graphs.trace_dependencies(noop1, (0.0, 2, 3))
    assert set(dependencies[0]) == set()

    # Avoid spurious hit
    def false_noop(x, y):
        return x + y

    dependencies = wn.causal_graphs.trace_dependencies(false_noop, (0.0, 1.0))
    assert set(dependencies[0]) == set([0, 1])

    dependencies = wn.causal_graphs.trace_dependencies(false_noop, (0.0, 0.0))
    assert set(dependencies[0]) == set([0, 1])


def test_dependency_tracing_numpy():
    """Test for tracing dependencies with numpy operations."""

    def func(a, b, c):
        """Simple matrix multiplication test."""
        vec = np.array([a, b, c])
        mat = np.array([[1.0, 0.0, 0.0], [1.0, 1.0, 1.0], [1.0, 0, 1.0]])

        output = np.dot(mat, vec)
        return output

    dependencies = wn.causal_graphs.trace_dependencies(func, (3, 5, 7))
    assert set(dependencies[0]) == set([0])
    assert set(dependencies[1]) == set([0, 1, 2])
    assert set(dependencies[2]) == set([0, 2])


def test_error_handling():
    """Test dependency tracing correctly handles multiple datatypes."""

    def func(a, b):
        return a + b

    # Floats should succeed
    deps = wn.causal_graphs.trace_dependencies(func, (1.0, 0.2))
    assert set(deps[0]) == set([0, 1])

    # Int should be cast and work
    deps = wn.causal_graphs.trace_dependencies(func, (1.0, 2))
    assert set(deps[0]) == set([0, 1])

    a = np.random.rand(3, 3).astype(np.float32)
    b = np.random.rand(3, 3).astype(np.int32)
    deps = wn.causal_graphs.trace_dependencies(func, (a, b))
    assert set(deps[0]) == set([0, 1])

    a = np.random.rand(3, 3).astype(np.int32)
    b = np.random.rand(3, 3).astype(np.int32)
    deps = wn.causal_graphs.trace_dependencies(func, (a, b))
    assert set(deps[0]) == set([0, 1])

    # Should raise for string arguments
    with pytest.raises(ValueError):
        wn.causal_graphs.trace_dependencies(func, (1.0, "str"))


###########################
# Test tracing the dynamics
###########################
@dataclasses.dataclass
class SimpleState(BaseState):
    """State for simple dynamics simulator."""

    x1: float = 1.0
    x2: float = 2.0
    x3: float = 3.0


@dataclasses.dataclass
class SimpleConfig(BaseConfig):
    """Config for simple dynamics simulator."""

    param: float = 2.0
    start_time: float = 0.0
    end_time: float = 5.0


class SimpleIntervention(BaseIntervention):
    """Change param at the given time."""

    def __init__(self, time=5, **kwargs):
        super(SimpleIntervention, self).__init__(SimpleConfig, time, **kwargs)


def simple_dynamics(state, time, config):
    """Toy dynamical system simulator for unit tests.

    The simulator cycles through three connectivity patterns, depending on the
    time.
    """
    x1, x2, x3 = state
    connectivity = math.floor(time) % 3
    if connectivity == 0:
        # Fully connected
        dx1 = x1 + x2 + x3
        dx2 = config.param * x1 + x2 + x3
        dx3 = x1 + x2 + x3
    elif connectivity == 1:
        # Diagonal
        dx1 = x1
        dx2 = x2
        dx3 = x3
    else:
        # Only x2 depends on input
        dx1 = config.param
        dx2 = x1 + x2 + x3
        dx3 = 0.0

    return [dx1, dx2, dx3]


class SimpleSimulator:
    """Mock of a module."""

    def __init__(self):
        self.dynamics = simple_dynamics
        self.SUPPORTS_CAUSAL_GRAPHS = True

    @staticmethod
    def simulate(initial_state, config, intervention=None, seed=None):
        """Simulate function for simple dynamics."""
        states = [initial_state]
        state = initial_state.values()
        times = [0]
        for time in range(int(config.start_time), int(config.end_time)):
            if intervention and time >= intervention.time:
                cfg = config.update(intervention)
            else:
                cfg = config
            delta = np.array(simple_dynamics(state, time, cfg))
            state += delta
            states.append(SimpleState(*np.copy(state)))
            times.append(time + 1)
        return wn.dynamics.Run(states=states, times=times)


def test_dynamics_tracer():
    """Sanity check for dynamics tracer."""
    tracer = wn.causal_graphs.trace_dynamics(simple_dynamics)
    for time in range(10):
        state = SimpleState()
        config = SimpleConfig()
        dependencies = tracer(state, time, config)

        if time % 3 == 0:
            assert set(dependencies["x1"]["states"]) == set(["x1", "x2", "x3"])
            assert set(dependencies["x1"]["configs"]) == set([])
            assert set(dependencies["x2"]["states"]) == set(["x1", "x2", "x3"])
            assert set(dependencies["x2"]["configs"]) == set(["param"])
            assert set(dependencies["x3"]["states"]) == set(["x1", "x2", "x3"])
            assert set(dependencies["x3"]["configs"]) == set([])
        elif time % 3 == 1:
            assert set(dependencies["x1"]["states"]) == set(["x1"])
            assert set(dependencies["x1"]["configs"]) == set([])
            assert set(dependencies["x2"]["states"]) == set(["x2"])
            assert set(dependencies["x2"]["configs"]) == set([])
            assert set(dependencies["x3"]["states"]) == set(["x3"])
            assert set(dependencies["x3"]["configs"]) == set([])
        else:
            assert set(dependencies["x1"]["states"]) == set([])
            assert set(dependencies["x1"]["configs"]) == set(["param"])
            assert set(dependencies["x2"]["states"]) == set(["x1", "x2", "x3"])
            assert set(dependencies["x2"]["configs"]) == set([])
            assert set(dependencies["x3"]["states"]) == set([])
            assert set(dependencies["x3"]["configs"]) == set([])


def test_ate_causal_graph_builder():
    """Test the causal_graphs=True argument of dynamics experiments."""
    # We need to import numpy like this to enable tracing!
    import whynot.traceable_numpy as wnp

    def covariate_builder(run):
        # Get x1, x2, and x3 from different time steps.
        return wnp.array([run[0].x1, run[2].x2, run[3].x3])

    def outcome_extractor(run):
        # Run the sum of states in the last time step
        return wnp.sum(run.states[-1].values())

    def soft_threshold(x, tau, r=200):
        return 1.0 / (wnp.exp(tau * r - r * x) + 1)

    def propensity_scores(untreated_run, intervention):
        # Treat the run if x1 + x2 above some "soft-threshold"
        # at the intervention time
        run = untreated_run
        return 1.0 - 0.9 * soft_threshold(run[0].x1 + run[4].x2, tau=4)

    intervention = SimpleIntervention(time=3, param=1.0)
    experiment = wn.DynamicsExperiment(
        name="test_experiment.",
        description="testing_ate",
        simulator=SimpleSimulator(),
        simulator_config=SimpleConfig(param=2.0, end_time=6),
        intervention=intervention,
        state_sampler=lambda: SimpleState(),
        propensity_scorer=propensity_scores,
        outcome_extractor=outcome_extractor,
        covariate_builder=covariate_builder,
    )

    dset = experiment.run(num_samples=10, causal_graph=True)
    graph = dset.causal_graph
    times = list(range(0, 7))
    state_names = SimpleState.variable_names()
    config_names = SimpleConfig.parameter_names()

    # First check the graph contains exactly the nodes we expect
    nodes = copy.deepcopy(list(graph.nodes))
    nodes.remove("Treatment")
    nodes.remove("Outcome")
    for time in times:
        for state in state_names:
            nodes.remove(f"{state}_{time}")
        for config_name in config_names:
            nodes.remove(f"PARAM:{config_name}_{time}")
    assert len(nodes) == 0

    # Check the covariates are correct
    assert set(graph.graph["covariate_names"]) == set(["x1_0", "x2_2", "x3_3"])

    # Now, check the graph contains the edges we expect
    edges = copy.deepcopy(list(graph.edges))
    # Check outcome depends on all states at the last time
    for outcome_dep in [f"{name}_{times[-1]}" for name in state_names]:
        edges.remove((outcome_dep, "Outcome"))
    # Check incoming treatment edges are correct
    edges.remove(("x1_0", "Treatment"))
    edges.remove(("x2_4", "Treatment"))
    # Check outgoing treatment edges
    for time in times:
        if time >= intervention.time:
            edges.remove(("Treatment", f"PARAM:param_{time}"))

    # Now check edges match the simulator dynamics.
    for time in times[:-1]:
        if time % 3 == 0:
            start_nodes = [f"{name}_{time}" for name in state_names]
            end_nodes = [f"{name}_{time+1}" for name in state_names]
            for edge in itertools.product(start_nodes, end_nodes):
                edges.remove(edge)
            edges.remove((f"PARAM:param_{time}", f"x2_{time+1}"))
        elif time % 3 == 1:
            for name in state_names:
                edges.remove((f"{name}_{time}", f"{name}_{time+1}"))
        else:
            edges.remove((f"PARAM:param_{time}", f"x1_{time+1}"))
            for name in state_names:
                edges.remove((f"{name}_{time}", f"x2_{time+1}"))

    assert len(edges) == 0



"""Basic experiments for opioid simulator."""
import numpy as np

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import opioid

__all__ = [
    "get_experiments",
    "RCT",
    "Confounding",
    "UnobservedConfounding",
    "Mediation",
]


def get_experiments():
    """Return all of the experiments for opioid simulator."""
    return [RCT, Confounding, UnobservedConfounding, Mediation]


def sample_initial_states(rng):
    """Sample initial states based on estimated uncertainty in Chen paper."""

    def sample(mean, std_dev):
        return mean + std_dev * rng.randn()

    nonmedical_users = sample(10029859, 100 * 329807)
    oud_users = sample(1369218, 100 * 116347)
    illicit_users = sample(328731, 100 * 60446)

    return opioid.State(
        nonmedical_users=nonmedical_users,
        oud_users=oud_users,
        illicit_users=illicit_users,
    )


@parameter(
    name="nonmedical_incidence_delta",
    default=-0.113,
    values=[-0.073, -0.113],
    description="Percent decrease in new nonmedical users of prescription opioids.",
)
def opioid_intervention(nonmedical_incidence_delta):
    """Return treatment config. Changes incidence of nonmedical abuse."""
    return opioid.Intervention(
        time=2015, nonmedical_incidence=nonmedical_incidence_delta
    )


def overdose_deaths(config, intervention, run):
    """Compute overdose deaths from the decade after intervention."""
    return opioid.simulator.compute_overdose_deaths(
        run, intervention.time, intervention.time + 10, config, intervention
    )


########################
# Randomized Experiment
#######################
#: RCT on the opioid simulator to see how reducing nonmedical opioid use reduces overdose deaths.
RCT = DynamicsExperiment(
    name="opioid_rct",
    description=(
        "Randomized experiment reducing nonmedical incidence of " "opioid use in 2015."
    ),
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=0.5,
    outcome_extractor=overdose_deaths,
    covariate_builder=lambda run: run.initial_state.values(),
)


########################
# Confounded experiments
########################
@parameter(
    name="propensity",
    default=0.9,
    values=[0.5, 0.6, 0.7, 0.9, 0.99],
    description="Probability of treatment assignment in positive group.",
)
def confounded_propensity_scores(config, intervention, untreated_runs, propensity):
    """Rollouts with top 10% of illicit deaths in 2015 more likely to receive treatment."""
    num_samples = len(untreated_runs)

    # More likely to assign treatment when illicit overdose
    # deaths are high in 2015
    illicit_deaths = [
        opioid.simulator.compute_illicit_deaths(run, 2015, config, intervention)
        for run in untreated_runs
    ]

    # Get top 10% of cities based on illicit overdose deaths
    sorted_idxs = sorted(range(num_samples), key=lambda idx: illicit_deaths[idx])
    top_death_idxs = sorted_idxs[int(num_samples * 0.9) :]

    propensity_scores = (1.0 - propensity) * np.ones(num_samples)
    propensity_scores[top_death_idxs] = propensity

    return propensity_scores


# pylint: disable-msg=invalid-name
#: Observational version of RCT with confounding on observed nonmedical use overdose deaths.
Confounding = DynamicsExperiment(
    name="confounded_opioid_experiment",
    description="Opioid confounding by treating runs with high numbers of nonmedical use deaths.",
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=overdose_deaths,
    covariate_builder=lambda intervention, run: run[intervention.time].values(),
)


# pylint: disable-msg=invalid-name
#: Observational version of RCT with unobserved confounding on nonmedical use overdose deaths.
UnobservedConfounding = DynamicsExperiment(
    name="unobserved_confounding_opioid_experiment",
    description=(
        "Opioid unobserved confounding by treating runs with high "
        "numbers of illicit deaths and omitting this variables."
    ),
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=overdose_deaths,
    covariate_builder=lambda intervention, run: np.array(
        [run[intervention.time].oud_users, run[intervention.time].nonmedical_users]
    ),
)


########################
# Mediation experiments
########################
@parameter(
    name="mediation_year",
    default=2020,
    values=[2017, 2019, 2020, 2021, 2024],
    description="Year of mediating variables--proxy for mediation strength." "",
)
def mediation_covariates(run, mediation_year):
    """Construct covariates for mediation experiment.

    Mediators are future state variables. Block backdoor paths with
    state variables in intervention year.
    """
    confounders = run[2015].values()
    mediation_state = run[mediation_year]
    mediators = np.array([mediation_state.nonmedical_users, mediation_state.oud_users])
    return np.concatenate([confounders, mediators])


# pylint: disable-msg=invalid-name
#: Observational version of RCT with mediation induced by observing future states.
Mediation = DynamicsExperiment(
    name="opioid_mediation",
    description="Mediation on opioid experiment by including future state.",
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=overdose_deaths,
    covariate_builder=mediation_covariates,
)

if __name__ == "__main__":
    RCT.run(num_samples=10)



"""World3 initialization."""

from whynot.simulators.world3.simulator import Config, Intervention, simulate, State
from whynot.simulators.world3.experiments import *
from whynot.simulators.world3.environments import *

SUPPORTS_CAUSAL_GRAPHS = False



"""Infrastructure for dynamical system simulators and experiments."""
import copy
import dataclasses
import inspect

import numpy as np
from tqdm.auto import tqdm

from whynot.framework import Dataset, extract_params, ParameterCollection
from whynot import causal_graphs, utils


class BaseConfig:
    # pylint: disable-msg=too-few-public-methods
    """Parameters for the simulation dynamics."""

    # Start and end time.
    start_time: float = 0
    end_time: float = 0
    # Time elapsed between measurements of the system.
    delta_t: float = 1.0

    def update(self, intervention):
        """Generate a new config after applying the intervention."""
        return dataclasses.replace(self, **intervention.updates)

    @classmethod
    def parameter_names(cls):
        """Return the parameters of the simulator, excluding start, end, and delta_t."""
        excluded = ["start_time", "end_time", "delta_t"]
        return [f.name for f in dataclasses.fields(cls) if f.name not in excluded]


@dataclasses.dataclass
class BaseState:
    # pylint: disable-msg=too-few-public-methods
    """State of the simulator."""

    @classmethod
    def num_variables(cls):
        """Count the number of variables."""
        return len(dataclasses.fields(cls))

    @classmethod
    def variable_names(cls):
        """List the variable names."""
        return [f.name for f in dataclasses.fields(cls)]

    def values(self):
        """Return the state as a numpy array."""
        # This ensures the values are shallow copied, unlike
        # dataclasses.astuple, which is necessary for graph tracing.
        return np.array([getattr(self, name) for name in self.variable_names()])


class BaseIntervention:
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention to a config."""

    def __init__(self, config_class, time, **kwargs):
        """Specify an intervention in the dynamical system.

        Parameters
        ----------
            time: int
                Time of the intervention.
            config_class:
                The Config class, a child class of dataclass.
            kwargs: dict
                Only valid keyword arguments are parameters of the config class.

        """
        self.time = time
        config_args = set(f.name for f in dataclasses.fields(config_class))
        for arg in kwargs:
            if arg not in config_args:
                raise TypeError(f"__init__() got an unexpected keyword argument {arg}!")
        self.updates = kwargs


@dataclasses.dataclass
class Run:
    r"""Encapsulate a trajectory from a dynamical system simulator.

    Attributes
    ----------
        states:
            Sequence of states :math:`x_{t_1}, x_{t_2}, \dots` produced by the
            simulator.
        times:
            Sequence of sampled times :math:`{t_1},{t_2}, \dots` at which states
            are recorded.

    Examples
    --------
    >>> state_at_year_closest_to_2015 = run[2015]

    """

    states: list
    times: list

    def __post_init__(self):
        """Error checking called after the constructor."""
        if len(self.states) != len(self.times):
            msg = "Input states and times must be the same length!  {} \neq {}"
            raise ValueError(msg.format(len(self.states), len(self.times)))

    def __getitem__(self, time):
        """Return the state closest to the given time."""
        time_index = np.argmin(np.abs(np.array(self.times) - time))
        return self.states[time_index]

    @property
    def initial_state(self):
        """Return initial state of the run."""
        return self.states[0]

    @property
    def initial_time(self):
        """Return the initial time of the run."""
        return self.times[0]

    @property
    def final_state(self):
        """Return the final state of the run."""
        return self.states[-1]

    @property
    def final_time(self):
        """Return the final time of the run."""
        return self.times[-1]


class DynamicsExperiment:
    """Encapsulate a causal experiment on a system dynamics simulator."""

    # pylint:disable-msg=too-many-arguments
    def __init__(
        self,
        name,
        description,
        simulator_config,
        intervention,
        state_sampler,
        simulator,
        propensity_scorer,
        outcome_extractor,
        covariate_builder,
    ):
        """Create an experiment based on system dynamics simulator.

        Parameters
        ----------
            name: str
                Name of the experiment
            description: str
                Short description of the experiment
            simulator:
                WhyNot systems dynamics simulator
            simulator_config: `Config`
                Instantiated simulator `Config` for all runs
            intervention: `Intervention`
                Instantiated simulator `Intervention` for treated runs
            state_sampler:
                Function to sample initial simulator `State` (s)
            propensity_score:
                Either float or function to generate propensity scores for a set of rollouts
            outcome_extractor:
                Function to extract outcome measurement from rollouts
            covariate_builder:
                Function to build covariates from rollouts.

        All of ``simulator_config``, ``intervention``, ``state_sampler``,
        ``propensity_scorer``, ``outcome_extractor``, and ``covariate_builder``
        support a rich set of patterns. We describe the expected API below.

        Note: If you wish to use causal graph construction and want to use numpy
        in any of the user-defined functions, please import the thinly wrapped version
        of numpy `whynot.traceable_numpy` and use this for all of the numpy
        calls, e.g.

        .. code-block:: python

            import whynot.traceable_numpy as np

            def propensity_scorer(run):
                return run[0].covariate_1 / np.max(run[0].values())


        Examples
        --------
        First, ``simulator_config`` can either be a constant ``Config`` object, or a
        parameterized function that returns a ``Config``. For
        instance:

        .. code-block:: python

            # simulator_config can be a constant Config object,
            DynamicsExperiment(
                ...
                simulator_config=Config(123),
                ...)
            
            # Or simulator_config can be a (parameterized) function that
            # returns a config.
            @parameter(name="p", default=0.2)
            def config(p):
                return Config(1234, parameter=p)

            DynamicsExperiment(
                ...
                simulator_config=config,
                ...)

        Similarly, ``intervention`` can either be a constant ``Intervention``
        object, or a (parameterized) function that returns an
        ``Intervention``.

        .. code-block:: python

                DynamicsExperiment(
                    ...
                    intervention=Intervention(year=92),
                    ...)

                @parameter(name="p", default=0.2)
                def intervention(p):
                    return Intervention(year=92, parameter=p)

                DynamicsExperiment(
                    ...
                    intervention=intervention,
                    ...)

        The ``state_sampler`` generates a sequence of initial states. If the
        sampler uses randomness, it must include ``rng`` in the signature and use
        it as the sole source of randomness.  This is to ensure that
        DynamicsExperiments can be deterministic if desired.
        
        If ``num_samples`` is in the signature, then the ``state_sampler`` must
        returns a sequence of ``num_samples`` states.  Otherwise, it must
        returns a single state.

        .. code-block:: python

            def state_sample(rng):
                # Return a single random sample, use randomness in rng
                return State(rng.rand())

            def state_sampler(rng, num_samples):
                # Return a list of initial states
                return [State(rng.rand()) for _ in range(num_samples)]

        The ``propensity_scorer`` produces, for each run, the probability of
        being assigned to treatment in the observational dataset. The
        ``propensity_scorer`` can be a constant, in which case the probability
        of treatment is uniform for all units. Otherwise, if
        ``propensity_scorer`` is a function, it either produces a propensity
        score for each run separately or all together (for correlated treatment
        assignment).

        For individual runs, the function may include one or both of
        ``treated_run`` and ``untreated_run`` in the signature.  For a whole
        population, use ``untreated_runs`` or ``treated_runs`` in the function
        signature.  Both ``config`` and ``intervention`` can be passed as
        argument to access the ``simulator_config`` and ``intervention`` for the
        given experiment.

        .. code-block:: python

            # self.propensity_scorer can be a constant
            DynamicsExperiment(
                ...
                propensity_scorer=0.8),
                ...)

            ##
            # Compute treatment assignment probability for each run separately.
            ##
            def propensity_scorer(untreated_run):
                # Return propensity score for a single run, based on the
                # untreated rollout.
                return 0.5 if untreated_run[20].values()[0] > 0.2 else 0.1

            def propensity_scorer(untreated_run, config, intervention):
                # You can pass the config and intervention
                intervention_covariate = untreated_run[intervention.year].values()[0]
                threshold = config.threshold
                return 1.0 if intervention_covariate > threshold else 0.0

            def propensity_scorer(treated_run):
                # Return propensity score for a single run, based on treated run
                return 0.5 if treated_run[20].values()[0] > 0.2 else 0.1

            def propensity_scorer(treated_run, untreated_run, config):
                # Assign treatment based on runs with both settings.
                if treated_run[10].values()[0] > 5 and untreated_run[10].values() < 10:
                    return config.propensity
                return 1.0 - config.propensity

            ##
            # Compute treatment assignment probability for all runs
            # simultaneously. This allows for correlated propensity scores.
            ##
            def propensity_scorer(untreated_runs):
                # Return propensity score for all runs, based on untreated run.
                # Only treat the top 10% of runs.
                covariates = [run.final_state.values()[0] for run in untreated_runs]
                top10 = np.argsort(covariates)[-10:]
                return [0.9 if idx in top10 else 0.1 for idx in range(len(untreated_runs))]

            etc..

        The ``outcome_extractor`` returns, for a given run, the outcome
        :math:`Y`. Both ``config`` and ``intervention`` can be passed as
        arguments.  For instance,

        .. code-block:: python

            def outcome_extractor(run):
                # Return the first state coordinate at time step 100
                return run[100].values()[0]

            def outcome_extractor(run, config, intervention):
                # You can also specify one (or both) of config/intervention
                # First state coordinate 20 steps after intervention.
                return run[intervention.time + 20].values()[0]

        Finally, the ``covariate_builder`` returns, for a given run, the
        observed covariates :math:`X`. If the signature includes ``runs``, the
        method returns covariates for the entire sequence of runs. Otherwise, it
        produces covariates for a single run. Both ``config`` and
        ``intervention`` can be passed.

        .. code-block:: python

            def covariate_builder(run):
                # Functions with the argument `run` return covariates for a
                # single rollout.
                return run.initial_state.values()

            def covariate_builder(runs):
                # Functions with the argument `runs` return covariates for the
                # entire sequence of runs.
                return np.array([run.initial_state.value() for run in runs])

            def covariate_builder(config, intervention, run):
                # Can optionally specify `config` or `intervention` in both cases.
                return run[intervention.year].values()

        """
        self.name = name
        self.description = description
        self.simulator = simulator

        if not callable(simulator_config):
            self.simulator_config = lambda: simulator_config
        else:
            self.simulator_config = simulator_config

        if not callable(intervention):
            self.intervention = lambda: intervention
        else:
            self.intervention = intervention

        self.state_sampler = state_sampler

        if not callable(propensity_scorer):
            if isinstance(propensity_scorer, float) and 0.0 <= propensity_scorer <= 1.0:
                self.propensity_scorer = lambda: propensity_scorer
            else:
                msg = "If propensity scorer is not callable, it must be a float in [0., 1.]"
                raise ValueError(msg)
        else:
            self.propensity_scorer = propensity_scorer

        self.outcome_extractor = outcome_extractor
        self.covariate_builder = covariate_builder
        self.parameter_collection = self.get_parameters()

    def get_parameters(self):
        """Inspect provided methods and gather parameters.

        Returns
        -------
            params: `whynot.framework.ParameterCollection`
                Collection of all of the parameters specified in the experiment.

        """
        params = ParameterCollection()

        config_args = []
        params += extract_params(self.simulator_config, config_args)
        params += extract_params(self.intervention, config_args)

        initial_state_args = ["rng", "num_samples"]
        params += extract_params(self.state_sampler, initial_state_args)

        propensity_args = [
            "config",
            "intervention",
            "untreated_run",
            "untreated_runs",
            "treated_run",
            "treated_runs",
        ]
        params += extract_params(self.propensity_scorer, propensity_args)

        outcome_args = ["config", "intervention", "run"]
        params += extract_params(self.outcome_extractor, outcome_args)

        covariate_args = ["config", "intervention", "run", "runs"]
        params += extract_params(self.covariate_builder, covariate_args)

        return params

    @staticmethod
    def _get_args(func):
        """Return the arguments to the function."""
        return inspect.signature(func).parameters

    @staticmethod
    def _run_dynamics_simulator(simulate, config, intervention, initial_state, seed):
        """Run the simulation with a common initial state for both treated and untreated.

        Use the same initial seed for each run, and return the outcomes of both runs.
        This function is a staticmethod since self is not pickable when the
        class has lambda functions as attributes, e.g. the propensity scorer.

        Parameters
        ----------
            simulate: function
                Entry point to the dynamical systems simulator.
            config:   `Config`
                Configuration object for the simulator, used for both treated and untreated.
            intervention: `Intervention`
                Intervention object for the simulator, used for the treated group.
            initial_state: `State`
                State object used to initial both treated and untreated runs of the simulator.
            seed: int
                Random seed used to initialize both runs of the simulator.

        Returns
        -------
            (untreated_run, treated_run): `Run`
                Tuple containing runs of the simulator with and without treatment.

        """
        # Work with copies of the initial state to avoid issues where the
        # simulator had side effects on the state.
        untreated_run = simulate(
            copy.deepcopy(initial_state), config=config, intervention=None, seed=seed
        )
        treated_run = simulate(
            copy.deepcopy(initial_state),
            config=config,
            intervention=intervention,
            seed=seed,
        )
        return untreated_run, treated_run

    def _run_all_simulations(
        self, initial_states, config, intervention, rng, parallelize, show_progress
    ):
        """Run simulation for treated and untreated groups for each initial state.

        Parameters
        ----------
            initial_state: list
                List of initial state objects
            config: `Config`
                Configuration object for all runs of the simulator.
            intervention: `Intervention`
                Intervention object for treated runs of the simulator.
            rng: `np.random.RandomState`
                Source of randomness. Used to choose seeds for different simulator runs.
            parallelize: bool
                Whether or not to execute runs of the simulator in parallel or sequentially.
            show_progress: bool
                Whether or not to display a progress bar for simulator execution.

        Returns
        -------
            untreated_runs: `list[Run]`
                List of simulator runs in untreated group, one for each initial state.
            treated_runs: `list[Run]`
                List of simulator runs in treated group, one for each initial state.

        """
        parallel_args = []
        for state in initial_states:
            seed = rng.randint(0, 2 ** 32 - 1)
            parallel_args.append(
                (self.simulator.simulate, config, intervention, state, seed)
            )

        if parallelize:
            runs = utils.parallelize(
                self._run_dynamics_simulator, parallel_args, show_progress=show_progress
            )
        elif show_progress:
            runs = [self._run_dynamics_simulator(*args) for args in tqdm(parallel_args)]
        else:
            runs = [self._run_dynamics_simulator(*args) for args in parallel_args]

        untreated_runs, treated_runs = list(zip(*runs))

        return untreated_runs, treated_runs

    def construct_covariates(
        self, config, intervention, untreated_runs, treated_runs, treatments, params
    ):
        """Build the observational dataset after running simulator.

        Parameters
        ----------
            config: `Config`
                Config object used when running the simulator.
            intervention: `Intervention`
                Intervention object used when running the simulator with treatment.
            untreated_runs: `list[Run]`
                List of all of the simulator runs in untreated condition.
            treated_runs: `list[Run]`
                List of all of the simulator runs with treatment applied.
            treatments: `list[bool]`
                Whether or not a particular run was assigned to treatment or untreated.
            params: dict
                Dictionary mapping parameter names to assigned values.

        Returns
        -------
            covariates: `np.ndarray`
                Numpy array of shape [num_samples, num_features] containing the
                covariates for the observational dataset.

        """
        observed_runs = []
        for treatment, untreated_run, treated_run in zip(
            treatments, untreated_runs, treated_runs
        ):
            run = treated_run if treatment else untreated_run
            observed_runs.append(run)

        covariate_builder_args = self._get_args(self.covariate_builder)
        kwargs = {}
        for arg in covariate_builder_args:
            if arg in params:
                kwargs[arg] = params[arg]

        if "config" in covariate_builder_args:
            kwargs["config"] = config

        if "intervention" in covariate_builder_args:
            kwargs["intervention"] = intervention

        # Generate all of the covariates in batch
        if "runs" in covariate_builder_args:
            dataset = self.covariate_builder(runs=observed_runs, **kwargs)
        else:
            # Generate the covariates one by one

            dataset = np.array(
                [
                    self.covariate_builder(run=observed_run, **kwargs)
                    for observed_run in observed_runs
                ]
            )

        if dataset.ndim == 1:
            dataset = np.expand_dims(dataset, axis=1)
        return dataset

    def get_outcomes(self, config, intervention, untreated_runs, treated_runs, params):
        """Return outcomes for both treated and untreated runs.

        Parameters
        ----------
            config: `Config`
                Simulator configuration object used to generate the runs.
            intervention: `Intervention`
                Simulator intervention used to specify treated runs.
            untreated_runs:   `list[Run]`
                List of simulator runs without intervention.
            treated_runs: `list[Run]`
                List of simulator runs with intervention.
            params:
                Dictionary mapping parameter names to arguments.

        Returns
        -------
            untreated_outcomes: `np.ndarray`
                Array of shape [num_samples] with untreated outcome :math:`Y_i(0)` for each i.
            treated_outcomes: `np.ndarray`
                Array of shape [num_samples] with treated outcome :math:`Y_i(1)` for each i.

        """
        extractor_args = self._get_args(self.outcome_extractor)
        kwargs = {}
        for arg in extractor_args:
            if arg in params:
                kwargs[arg] = params[arg]
        if "config" in extractor_args:
            kwargs["config"] = config
        if "intervention" in extractor_args:
            kwargs["intervention"] = intervention

        def get_outcome(run):
            return self.outcome_extractor(run=run, **kwargs)

        untreated_outcomes = np.array([get_outcome(run) for run in untreated_runs])
        treated_outcomes = np.array([get_outcome(run) for run in treated_runs])

        return untreated_outcomes, treated_outcomes

    def get_propensity_scores(
        self, config, intervention, untreated_runs, treated_runs, params
    ):
        """Return probability of treatment for a set of runs.

        Parameters
        ----------
            config: `Config`
                Simulator configuration object used to generate the runs.
            intervention: `Intervention`
                Simulator intervention used to specify treated runs.
            untreated_runs:   `list[Run]`
                List of simulator runs without intervention.
            treated_runs: `list[Run]`
                List of simulator runs with intervention.
            params:
                Dictionary mapping parameter names to arguments.

        Returns
        -------
            propensity_score: `np.ndarray`
                Array of shape [num_samples] with probability of treatment for each unit i.

        """
        # pylint:disable-msg=too-many-branches
        if not callable(self.propensity_scorer):
            if isinstance(self.propensity_scorer, float):
                uniform_propensity = self.propensity_scorer
                if 0.0 <= uniform_propensity <= 1.0:
                    return uniform_propensity * np.ones(len(untreated_runs))

            raise ValueError(
                "If propensity scorer is not callable, it must be a float in [0., 1.]"
            )

        scorer_args = self._get_args(self.propensity_scorer)

        # Build kwargs for scorer based on passed parameters
        kwargs = {}
        for arg in scorer_args:
            if arg in params:
                kwargs[arg] = params[arg]

        if "config" in scorer_args:
            kwargs["config"] = config

        if "intervention" in scorer_args:
            kwargs["intervention"] = intervention

        # Check whether or not the propensity scorer generates all scores at once
        one_shot = False
        if "treated_runs" in scorer_args:
            one_shot = True
            kwargs["treated_runs"] = treated_runs
        if "untreated_runs" in scorer_args:
            one_shot = True
            kwargs["untreated_runs"] = untreated_runs

        if one_shot:
            return self.propensity_scorer(**kwargs)

        # If not, then we need to score each run separately.
        propensity_scores = []
        for untreated_run, treated_run in zip(untreated_runs, treated_runs):
            if "untreated_run" in scorer_args:
                kwargs["untreated_run"] = untreated_run
            if "treated_run" in scorer_args:
                kwargs["treated_run"] = treated_run
            score = self.propensity_scorer(**kwargs)
            propensity_scores.append(score)

        return np.array(propensity_scores)

    def sample_initial_states(self, rng, num_samples, params):
        """Sample initial states for simulator.

        Parameters
        ----------
            rng: `np.random.RandomState`
                Random number generator used for all experiment randomness
            num_samples: int
                Number of samples to use for the experiment.
            params:
                Dictionary mapping parameter names to assigned values.

        Returns
        -------
            initial_states: `list[States]`
                List of length `num_samples` of simulator initial `States.

        """
        state_sampler_args = self._get_args(self.state_sampler)

        # Build kwargs
        kwargs = {}
        for arg in state_sampler_args:
            if arg in params:
                kwargs[arg] = params[arg]

        if "rng" in state_sampler_args:
            kwargs["rng"] = rng

        if "num_samples" in state_sampler_args:
            return self.state_sampler(num_samples=num_samples, **kwargs)

        return [self.state_sampler(**kwargs) for _ in range(num_samples)]

    def get_config(self, params):
        """Return the simulator config.

        Parameters
        ----------
            params: dict
                Dictionary mapping experiment parameter names to assigned values

        Returns
        -------
            config: `Config`
                Simulator configuration object used for this experiment.

        """
        if not callable(self.simulator_config):
            return self.simulator_config

        untreated_args = self._get_args(self.simulator_config)
        kwargs = {}
        for arg in untreated_args:
            if arg in params:
                kwargs[arg] = params[arg]

        return self.simulator_config(**kwargs)

    def get_intervention(self, params):
        """Return the intervention corresponding to treatment.

        Parameters
        ----------
            params: dict
                Dictionary mapping experiment parameter names to assigned
                values.

        Returns
        -------
            intervention: `Intervention`
                Simulator Intervention object used for treated group.

        """
        if not callable(self.intervention):
            return self.intervention

        treated_args = self._get_args(self.intervention)
        kwargs = {}
        for arg in treated_args:
            if arg in params:
                kwargs[arg] = params[arg]

        return self.intervention(**kwargs)

    def construct_causal_graph(
        self, untreated_runs, treated_runs, config, intervention, run_parameters
    ):
        """Construct the causal graph associated with the experiment.

        This feature is still experimental and will likely have rough
        edges.

        Parameters
        ----------
            untreated_runs: list
                List of untreated Run objects generated by the experiment.
            treated_runs: list
                List of treated Run objects generated by the experiment.
            config: whynot.dynamics.BaseConfig
                Fully instantiated simulator config used for the experiment.
            intervention: whynot.dynamics.BaseIntervention
                Fully instantiated simulator intervention used for the experiment.
            run_parameters: dict
                User-specified parameters associated with the experiment.

        Returns
        -------
            ate_graph: networkx.DiGraph
                Causal graph corresponding to the experiment.

        """
        # Extract dependencies for each of the ATE components.
        with causal_graphs.trace_stack.new_trace() as trace:
            # All of the supported simulators dont have data dependent
            # control flow, so for now, tracing the simulator and user
            # functions only requires a single run.
            run = untreated_runs[0]

            # Wrap the run object in a box
            run, node_map = causal_graphs.run_to_box(run, trace)
            untreated_runs, treated_runs = [run], [run]

            # Trace the propensity scorer
            propensities = self.get_propensity_scores(
                config, intervention, untreated_runs, treated_runs, run_parameters
            )
            treatment_deps = causal_graphs.backtrack(propensities[0], node_map)

            # Trace the covariate extractor.
            covariates = self.construct_covariates(
                config,
                intervention,
                untreated_runs,
                treated_runs,
                np.array([0.0]),
                run_parameters,
            )
            covariate_deps = causal_graphs.backtrack(covariates[0], node_map)

            # Trace the outcome extractor
            untreated_outcomes, _ = self.get_outcomes(
                config, intervention, untreated_runs, treated_runs, run_parameters
            )
            outcome_deps = causal_graphs.backtrack(untreated_outcomes[0], node_map)

        # Unpack the dependencies from backtracking
        treatment_deps = [node.name for node in treatment_deps[0]]
        outcome_deps = [node.name for node in outcome_deps[0]]
        covariate_deps = dict(
            (cov, [node.name for node in deps]) for cov, deps in covariate_deps.items()
        )

        # Build the causal graph corresponding to the dynamics.
        ate_graph = causal_graphs.ate_graph_builder(
            self.simulator,
            run,
            config,
            intervention,
            treatment_deps,
            covariate_deps,
            outcome_deps,
        )

        return ate_graph

    def run(
        self,
        num_samples,
        seed=None,
        parallelize=True,
        show_progress=False,
        causal_graph=False,
        **parameter_args,
    ):
        # pylint:disable-msg=too-many-locals
        """Run a basic parameterized experiment on a dynamical system simulator.

        Parameters
        ----------
            num_samples: int
                Number of units to sample and simulate
            show_progress: bool
                Should progress bar be shown?
            parallelize: bool
                If true, the experiment class will run all of the simulations in
                parallel. Parallelization is performed with multiprocessing
                using a ProcessPool from the concurrent.futures module. 
            seed: int
                Random number generator seed used to set all internal randomness.
            causal_graph:
                Whether to attempt to build the causal graph. Currently, this is only
                supported on the hiv, lotka_volterra, and opioid simulators.
            **parameter_args:
                If the experiment is parameterized, additional arguments to
                select the parameters of a particular run.

        Returns
        -------
            dataset: whynot.framework.Dataset
                Dataset object encapsulating the covariates, treatment assignments, and
                outcomes observed in the experiment, as well as unit-level ground truth.
                The covariates are an array of size [num_samples, num_features] where
                num_features is determined by the covariate builder.

        """
        if causal_graph and not self.simulator.SUPPORTS_CAUSAL_GRAPHS:
            error_msg = (
                "This simulator does not currently support causal graph generation."
                "Rerun the command with causal_graphs=False to continue."
            )
            raise ValueError(error_msg)

        run_parameters = self.parameter_collection.project(parameter_args)

        rng = np.random.RandomState(seed)

        config = self.get_config(run_parameters)
        intervention = self.get_intervention(run_parameters)

        initial_states = self.sample_initial_states(rng, num_samples, run_parameters)

        # Run the model
        untreated_runs, treated_runs = self._run_all_simulations(
            initial_states, config, intervention, rng, parallelize, show_progress
        )

        # Assign treatment
        propensities = self.get_propensity_scores(
            config, intervention, untreated_runs, treated_runs, run_parameters
        )
        treatment = rng.uniform(size=propensities.shape) < propensities
        treatment = treatment.astype(np.int64)

        # Build observational dataset
        covariates = self.construct_covariates(
            config,
            intervention,
            untreated_runs,
            treated_runs,
            treatment,
            run_parameters,
        )
        untreated_outcomes, treated_outcomes = self.get_outcomes(
            config, intervention, untreated_runs, treated_runs, run_parameters
        )

        # Assign treatment
        outcomes = np.copy(untreated_outcomes)
        outcomes[treatment == 1] = treated_outcomes[treatment == 1]

        # Ground truth effect is Y(1) - Y(0)
        treatment_effects = treated_outcomes - untreated_outcomes

        if causal_graph:
            graph = self.construct_causal_graph(
                untreated_runs, treated_runs, config, intervention, run_parameters
            )
        else:
            graph = None

        return Dataset(
            covariates=covariates,
            treatments=treatment,
            outcomes=outcomes,
            true_effects=treatment_effects,
            causal_graph=graph,
        )

    def __repr__(self):
        """Display the experiment by it's name."""
        return self.name



"""Unit tests for opioid epidemic simulator."""
import whynot as wn


def test_config():
    """Ensure intervention update works as expected."""
    intervention = wn.opioid.Intervention(time=2021, nonmedical_incidence=-0.12)
    config = wn.opioid.Config()
    assert config.nonmedical_incidence.intervention_val == 0.0
    config = config.update(intervention)
    assert config.nonmedical_incidence.intervention_val == -0.12

    intervention = wn.opioid.Intervention(time=2021, illicit_incidence=1.2)
    config = wn.opioid.Config()
    assert config.illicit_incidence.intervention_val == 0.0
    config = config.update(intervention)
    assert config.illicit_incidence.intervention_val == 1.2



"""Causal experiments for the Schelling model."""
import copy
import dataclasses

import numpy as np
import pandas as pd

from whynot.framework import GenericExperiment, parameter
from whynot.simulators import schelling
import whynot.utils as utils

__all__ = ["get_experiments", "RCT"]


def get_experiments():
    """Return all experiments for Schelling."""
    return [RCT]


################
# RCT Experiment
################
def run_simulator(control_config, treatment_config, seed):
    """Run the simulator in both the control and treatment conditions using a common seed."""
    control_segregation = schelling.simulate(control_config, rollouts=5, seed=seed)
    treatment_segregation = schelling.simulate(treatment_config, rollouts=5, seed=seed)
    return control_segregation, treatment_segregation


@parameter(
    name="education_pc",
    default=0.3,
    values=np.arange(0.0, 0.99, 0.05),
    description="Random fraction of the population to educate",
)
@parameter(
    name="education_boost",
    default=-1,
    values=[-1, -2, -3, -4],
    description="How much education decreases homophily",
)
def run_schelling(
    education_pc,
    education_boost,
    num_samples=100,
    seed=None,
    show_progress=False,
    parallelize=True,
):
    """Run a basic RCT experiment on Schelling.

    Each unit in the experiment is a grid or "community" in the Schelling model.
    Treatment corresponds to randomly educating some fraction of the
    agents in the grid to decrease their homophiliy. The measure outcome is the
    fraction of segregated agents on the grids. In this experiment, treatment is
    randomly assigned.

    Parameters
    ----------
        education_pc: float
            What percentrage of the population of the treated units should be "educated"
        education_boost: int
            How much receiving "education" changes the homophily of an agent
            Values in [-1, -2, -3, -4]
        num_samples: int
            How many units to sample for the experiment
        seed: int
            (Optional) Seed global randomness for the experiment
        show_progress: bool
            (Optional) Whether or not to print a progress bar
        parallelize: bool
            (Optional) Whether or not to use parallelism during the experiment

    Returns
    -------
        covariates: np.ndarray
            Array of shape [num_samples, num_features], observed covariates for
            each unit.
        treatment: np.ndarray
            Array of shape [num_samples], treatment assignment for each unit
        outcomes: np.ndarray
            Array of shape [num_samples], observed outcome for each unit
        ground_truth: np.ndarray
            Array of shape [num_samples], unit level treatment effects

    """
    rng = np.random.RandomState(seed)

    configs = []
    for _ in range(num_samples):
        config = schelling.Config()
        config.height = 10
        config.width = 10
        config.homophily = 5
        config.density = rng.uniform(0.05, 0.6)
        config.minority_pc = rng.uniform(0.05, 0.45)
        config.education_boost = education_boost

        # By default, no education.
        config.education_pc = 0.0
        treatment_config = copy.deepcopy(config)
        # Treatment corresponds to education
        treatment_config.education_pc = education_pc

        seed = rng.randint(0, 99999)
        configs.append((config, treatment_config, seed))

    if parallelize:
        runs = utils.parallelize(run_simulator, configs, show_progress=show_progress)
    else:
        runs = [run_simulator(*args) for args in configs]

    control_outcomes = np.array([run[0] for run in runs])
    treatment_outcomes = np.array([run[1] for run in runs])

    # Potential confounders are density and minority pc
    dataframe = pd.DataFrame([dataclasses.asdict(config[0]) for config in configs])
    covariates = dataframe[["density", "minority_pc"]].values

    # First pass: Randomly assign treatment
    # We could only treat "cities" with high density and
    # high minority pc
    treatment = (rng.rand(num_samples) < 0.25).astype(np.int32)
    treatment_idxs = np.where(treatment == 1.0)[0]

    outcomes = np.copy(control_outcomes)
    outcomes[treatment_idxs] = treatment_outcomes[treatment_idxs]

    treatment_effects = treatment_outcomes - control_outcomes
    return (covariates, treatment, outcomes), treatment_effects


#: RCT experiment to understand the effect of homophily reduction on total segregation.
RCT = GenericExperiment(
    name="schelling_rct",
    description=(
        "Test the efficacy of an education program to reduce "
        "homophiliy of agents in the Schelling model."
    ),
    run_method=run_schelling,
)



"""Opioid initialization."""

from whynot.simulators.opioid.simulator import (
    Config,
    Intervention,
    dynamics,
    simulate,
    State,
)
from whynot.simulators.opioid.experiments import *
from whynot.simulators.opioid.environments import *

SUPPORTS_CAUSAL_GRAPHS = True



"""Simulator initialization."""

from whynot.simulators import civil_violence
from whynot.simulators import credit
from whynot.simulators import delayed_impact
from whynot.simulators import dice
from whynot.simulators import hiv
from whynot.simulators import incarceration
from whynot.simulators import lalonde
from whynot.simulators import lotka_volterra
from whynot.simulators import opioid
from whynot.simulators import schelling
from whynot.simulators import world2
from whynot.simulators import world3
from whynot.simulators import zika



"""whynot package initialization."""

__version__ = "0.11.1"
from whynot.algorithms import *
from whynot.simulators import *
from whynot import causal_graphs, dynamics, framework
from whynot.dynamics import (
    DynamicsExperiment,
    Run,
)
from whynot.framework import (
    Dataset,
    InferenceResult,
    parameter,
)
from whynot import utils



"""DICE initialization."""

from whynot.simulators.dice.simulator import Config, Intervention, simulate, State
from whynot.simulators.dice.experiments import *

# Currently, we cannot trace through the dynamics defined in Pyomo.
SUPPORTS_CAUSAL_GRAPHS = False



"""Implementation of Adams et al.'s ODE simulator for HIV treatment.

Adams, Brian Michael, et al. Dynamic multidrug therapies for HIV: Optimal and
STI control approaches. North Carolina State University. Center for Research in
Scientific Computation, 2004. APA.

https://pdfs.semanticscholar.org/c030/127238b1dbad2263fba6b64b5dec7c3ffa20.pdf
"""
import numpy as np

from whynot.gym import spaces
from whynot.gym.envs import ODEEnvBuilder, register
from whynot.simulators.hiv import Config, Intervention, simulate, State


def get_intervention(action, time):
    """Return the intervention in the simulator required to take action."""
    action_to_intervention_map = {
        0: Intervention(time=time, epsilon_1=0.0, epsilon_2=0.0),
        1: Intervention(time=time, epsilon_1=0.0, epsilon_2=0.3),
        2: Intervention(time=time, epsilon_1=0.7, epsilon_2=0.0),
        3: Intervention(time=time, epsilon_1=0.7, epsilon_2=0.3),
    }
    return action_to_intervention_map[action]


def get_reward(intervention, state):
    """Compute the reward based on the observed state and choosen intervention."""
    Q, R1, R2, S = 0.1, 20000.0, 20000.0, 1000.0
    reward = S * state.immune_response - Q * state.free_virus
    reward -= R1 * (intervention.updates["epsilon_1"] ** 2)
    reward -= R2 * (intervention.updates["epsilon_2"] ** 2)
    return reward


def observation_space():
    """Return observation space.
 
    The state is (uninfected_T1, infected_T1, uninfected_T2, infected_T2,
    free_virus, immune_response) in units (cells/ml, cells/ml, cells/ml,
    cells/ml, copies/ml, cells/ml).
    """
    state_dim = State.num_variables()
    state_space_low = np.zeros(state_dim)
    state_space_high = np.inf * np.ones(state_dim)
    return spaces.Box(state_space_low, state_space_high)


HivEnv = ODEEnvBuilder(
    simulate_fn=simulate,
    config=Config(),
    initial_state=State(),
    # In this environment there are 4 actions defined by
    # epsilon_1 = 0 or 0.7 and epsilon_2 = 0 or 0.3.
    action_space=spaces.Discrete(4),
    observation_space=observation_space(),
    timestep=1.0,
    intervention_fn=get_intervention,
    reward_fn=get_reward,
)

register(
    id="HIV-v0", entry_point=HivEnv, max_episode_steps=400, reward_threshold=1e10,
)



"""Implementation of Adams et al.'s ODE simulator for HIV treatment.

Adams, Brian Michael, et al. Dynamic multidrug therapies for HIV: Optimal and
STI control approaches. North Carolina State University. Center for Research in
Scientific Computation, 2004. APA.

https://pdfs.semanticscholar.org/c030/127238b1dbad2263fba6b64b5dec7c3ffa20.pdf
"""

import dataclasses
import numpy as np

from scipy.integrate import odeint

import whynot as wn
from whynot.dynamics import BaseConfig, BaseState, BaseIntervention


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Parameters for the simulation dynamics.

    Examples
    --------
    # Run the simulation for 200 days with infected cell death rate 0.3
    hiv.Config(duration=200, delta=0.3)

    """

    # pylint: disable-msg=invalid-name
    # pylint: disable-msg=too-many-instance-attributes
    #: Target cell type 1 production (source) rate
    lambda_1: float = 10000.0
    #: Target cell type 2 production (source) rate
    lambda_2: float = 31.98
    #: Target cell type 1 death rate
    d_1: float = 0.01
    #: Target cell type 2 death rate
    d_2: float = 0.01
    #: Population 1 infection rate
    k_1: float = 8e-7
    #: Population 2 infection rate
    k_2: float = 1e-4
    #: Treatment efficacy reduction in population 2
    f: float = 0.34
    #: Infected cell death rate
    delta: float = 0.7
    #: Immune-induced clearance rate for population 1
    m_1: float = 1e-5
    #: Immune-induced clearance rate for population 2
    m_2: float = 1e-5
    #: Virions produced per infected cell
    N_T: float = 100
    #: Virus natural death rate
    c: float = 13
    #: Average number virions infecting a type 1 cell
    rho_1: float = 1
    #: Average number virions infecting a type 2 cell
    rho_2: float = 1
    #: Immune effector production (source) rate
    lambda_E: float = 1
    #: Maximum birth rate for immune effectors
    b_E: float = 0.3
    #: Saturation constant for immune effector birth
    K_B: float = 100
    #: Maximum death rate for immune effectors
    d_E: float = 0.25
    #: Saturation constant for immune effector death
    K_D: float = 500
    #: Natural death rate for immune effectors
    delta_E: float = 0.1

    # Treatment regimes-- only consider constant
    # treatments for now. The paper also references
    # a constant "fully efficacious" treatment with
    # e_1 = 0.7 and e_2 = 0.3.
    #: Drug efficacy
    epsilon_1: float = 0.0
    #: Efficacy of protease inhibitors
    epsilon_2: float = 0.0

    #: Simulation start time (in day)
    start_time: float = 0
    #: Simulation end time (in days)
    end_time: float = 400
    #: How frequently to measure simulator state
    delta_t: float = 0.05
    #: solver relative tolerance
    rtol: float = 1e-6
    #: solver absolute tolerance
    atol: float = 1e-6


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """State of the HIV simulator.

    The default state corresponds to an early infection state, defined by Adams
    et al. The early infection state is designed based on an unstable uninfected
    steady state by 1) adding one virus particle per ml of blood plasma, and 2)
    adding low levels of infected T-cells.
    """

    # pylint: disable-msg=invalid-name
    #: Uninfected CD4+ T-lymphocytes (cells/ml)
    uninfected_T1: float = 1e6
    #: Infected CD4+ T-lymphocytes (cells/ml)
    infected_T1: float = 1e-4
    #: Uninfected macrophages (cells/ml)
    uninfected_T2: float = 3198
    #: Infected macrophages (cells/ml)
    infected_T2: float = 1e-4
    #: Free virus (copies/ml)
    free_virus: float = 1
    #: Immune response CTL E (cells/ml)
    immune_response: float = 10


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the HIV model.

    Examples
    --------
    >>> # Starting in step 100, set epsilon_1 to 0.7 (leaving other variables unchanged)
    >>> Intervention(time=100, epsilon_1=0.7)

    """

    def __init__(self, time=100, **kwargs):
        """Specify an intervention in the dynamical system.

        Parameters
        ----------
            time: int
                Time of the intervention (days)
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


def dynamics(state, time, config, intervention=None):
    """Update equations for the HIV simulaton.

    Parameters
    ----------
        state:  np.ndarray, list, or tuple
            State of the dynamics
        time:   float
        config: hiv.Config
            Simulator configuration object that determines the coefficients
        intervantion: hiv.Intervention
            Simulator intervention object that determines when/how to update the
            dynamics.

    Returns
    -------
        ds_dt: list
            Derivative of the dynamics with respect to time

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    # Keep notation roughly consistent with the Adams paper.
    # pylint: disable-msg=invalid-name
    (
        uninfected_T1,
        infected_T1,
        uninfected_T2,
        infected_T2,
        free_virus,
        immune_response,
    ) = state

    delta_uninfected_T1 = (
        config.lambda_1
        - config.d_1 * uninfected_T1
        - (1 - config.epsilon_1) * config.k_1 * free_virus * uninfected_T1
    )

    delta_infected_T1 = (
        (1 - config.epsilon_1) * config.k_1 * free_virus * uninfected_T1
        - config.delta * infected_T1
        - config.m_1 * immune_response * infected_T1
    )

    delta_uninfected_T2 = (
        config.lambda_2
        - config.d_2 * uninfected_T2
        - (1 - config.f * config.epsilon_1) * config.k_2 * free_virus * uninfected_T2
    )

    delta_infected_T2 = (
        (1 - config.f * config.epsilon_1) * config.k_2 * free_virus * uninfected_T2
        - config.delta * infected_T2
        - config.m_2 * immune_response * infected_T2
    )

    delta_virus = (
        (1 - config.epsilon_2) * config.N_T * config.delta * (infected_T1 + infected_T2)
        - config.c * free_virus
        - free_virus
        * (
            (1.0 - config.epsilon_1) * config.rho_1 * config.k_1 * uninfected_T1
            + (
                (1.0 - config.f * config.epsilon_1)
                * config.rho_2
                * config.k_2
                * uninfected_T2
            )
        )
    )

    delta_immune_response = (
        config.lambda_E
        + (
            (config.b_E * (infected_T1 + infected_T2))
            / (infected_T1 + infected_T2 + config.K_B)
        )
        * immune_response
        - (
            (config.d_E * (infected_T1 + infected_T2))
            / (infected_T1 + infected_T2 + config.K_D)
        )
        * immune_response
        - config.delta_E * immune_response
    )

    ds_dt = [
        delta_uninfected_T1,
        delta_infected_T1,
        delta_uninfected_T2,
        delta_infected_T2,
        delta_virus,
        delta_immune_response,
    ]
    return ds_dt


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Adams HIV simulator model.

    The simulation starts at initial_state at time 0, and evolves the state
    using dynamics whose parameters are specified in config.

    Parameters
    ----------
        initial_state:  `whynot.simulators.hiv.State`
            Initial State object, which is used as x_{t_0} for the simulator.
        config:  `whynot.simulators.hiv.Config`
            Config object that encapsulates the parameters that define the dynamics.
        intervention: `whynot.simulators.hiv.Intervention`
            Intervention object that specifies what, if any, intervention to perform.
        seed: int
            Seed to set internal randomness. The simulator is deterministic, so
            the seed parameter is ignored.

    Returns
    -------
        run: `whynot.dynamics.Run`
            Rollout of the model.

    """
    # Simulator is deterministic, so seed is ignored
    # pylint: disable-msg=unused-argument
    t_eval = np.arange(
        config.start_time, config.end_time + config.delta_t, config.delta_t
    )

    solution = odeint(
        dynamics,
        y0=dataclasses.astuple(initial_state),
        t=t_eval,
        args=(config, intervention),
        rtol=config.rtol,
        atol=config.atol,
    )

    states = [initial_state] + [State(*state) for state in solution[1:]]
    return wn.dynamics.Run(states=states, times=t_eval)


if __name__ == "__main__":
    print(simulate(State(), Config()))



"""Implementation of delayed impact lending simulator based on Liu et al.

Liu, L., Dean, S., Rolf, E., Simchowitz, M., & Hardt, M. (2018, July). Delayed
Impact of Fair Machine Learning. In International Conference on Machine
Learning. (https://arxiv.org/abs/1803.04383)
"""
import copy
import dataclasses
import os
from typing import Callable

import whynot as wn
import whynot.traceable_numpy as np
from whynot.dynamics import BaseConfig, BaseState, BaseIntervention
from whynot.simulators.delayed_impact.fico import get_data_args as get_FICO_data


#################################
# Globally accessible FICO params
#################################
DATAPATH = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")
INV_CDFS, LOAN_REPAY_PROBS, _, GROUP_SIZE_RATIO, _, _ = get_FICO_data(DATAPATH)


def default_credit_scorer(score):
    """Report the underlying score without modification."""
    return score


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Parameters for the simulation dynamics."""

    #: Maps the true credit score to the reported score
    credit_scorer: Callable = default_credit_scorer

    #: Lending threshold for group 0
    threshold_g0: float = 650
    #: Lending threshold for group 1
    threshold_g1: float = 650

    #: Bank repayment utility
    repayment_utility: float = 1.0
    #: Bank default utility
    default_utility: float = -4.0

    #: Applicant's score change after repayment
    repayment_score_change: float = 75
    #: Applican't score change after default
    default_score_change: float = -150

    #: Minimum credit score
    min_score: int = 350
    max_score: int = 800

    #: Simulation start step (in rounds)
    start_time: float = 0

    #: Simulation end step (in rounds)
    end_time: float = 1

    #: Simulator step size (Unused)
    delta_t: float = 1


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """State of the lending simulator."""

    #: Group membership (sensitive attribute) 0 or 1
    group: int = 0
    #: Agent credit score
    credit_score: int = 700
    #: Running total of the banks profit/loss for the agent
    profits: float = 0


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the lending model.

    Examples
    --------
    >>> # Change the group 0 threshold to 700
    >>> Intervention(time=0, threshold_g0=700)

    """

    def __init__(self, time=100, **kwargs):
        """Specify an intervention in the dynamical system.

        Parameters
        ----------
            time: int
                Time of the intervention (days)
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


def lending_policy(config, group, score):
    """Determine whether or not a bank gives a loan."""
    # P(T = 1 | X, A=j) = 1     if X >= tau_j
    #                     0     otherwise.
    return (score >= config.threshold_g0) ** (1 - group) * (
        score >= config.threshold_g1
    ) ** group


def determine_repayment(rng, group, score):
    """Determine whether or not the agent repays the loan."""
    repayment_rate = (
        LOAN_REPAY_PROBS[0](score) ** (1 - group) * LOAN_REPAY_PROBS[1](score) ** group
    )
    # Sample a Bernoulli with the Gumbel-max trick to permit causal graph
    # tracing
    uniform = rng.uniform()
    return (
        np.log(repayment_rate / (1.0 - repayment_rate))
        + np.log(uniform / (1.0 - uniform))
    ) > 0.0


def update_score(config, score, loan_approved, repaid):
    """Update the agent's credit score after a lending interaction."""
    score_change = (
        config.repayment_score_change ** repaid ** loan_approved
        * config.default_score_change ** (1 - repaid) ** loan_approved
        * 0.0 ** (1 - loan_approved)
    )
    new_score = score + score_change
    return np.minimum(np.maximum(new_score, config.min_score), config.max_score)


def update_profits(config, profits, loan_approved, repaid):
    """Update the running total bank profit for the individual."""
    profit_change = (
        config.repayment_utility ** repaid ** loan_approved
        * config.default_utility ** (1 - repaid) ** loan_approved
        * 0.0 ** (1 - loan_approved)
    )
    return profits + profit_change


def dynamics(state, time, config, intervention=None, rng=None):
    """Update equations for the lending simulaton.

    Performs one round of interaction between the agent (represented
    by the state) and the bank.

    Parameters
    ----------
        state: whynot.simulators.delayed_impact.State
            Agent state at the beginning of the interaction.
        time: int
            Current round of interaction
        config: whynot.simulators.delayed_impact.Config
            Configuration object controlling the interaction, e.g. lending
            threshold and credit scoring
        intervention: whynot.simulators.delayed_impact.Intervention
            Intervention object specifying when and how to update the dynamics.
        rng: np.RandomState
            Seed random number generator for all randomness (optional)

    Returns
    -------
        state: whynot.simulators.delayed_impact.State
            Agent state after one lending interaction.

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    if rng is None:
        rng = np.random.RandomState(None)

    group, score, individual_profits = state

    # Credit bureau measures the agent's score
    measured_score = config.credit_scorer(score)

    # Bank decides whether or not to extend the user a loan
    loan_approved = lending_policy(config, group, measured_score)

    # The user (potentially) repays the loan
    repaid = determine_repayment(rng, group, score)

    # The credit score updates in response
    new_score = update_score(config, score, loan_approved, repaid)

    new_profits = update_profits(config, individual_profits, loan_approved, repaid)

    return group, new_score, new_profits


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the lending simulator.

    The simulation starts at initial_state, representing an agent before
    interacting with the lending institution. The simulator evolves the agent
    state through (repeated) interaction between the agent and the lender.  The
    dynamics encapsulate how the lending decisions and policies effect both the
    agent and the lender's profit. The parameters of the dynamics, e.g. the
    lending thresholds or the repayment model, are specified in the Config.

    Parameters
    ----------
        initial_state:  `whynot.simulators.delayed_impact.State`
            Initial State object, which is used as x_{t_0} for the simulator.
        config:  `whynot.simulators.delayed_impact.Config`
            Config object that encapsulates the parameters that define the dynamics.
        intervention: `whynot.simulators.delayed_impact.Intervention`
            Intervention object that specifies what, if any, intervention to perform.
        seed: int
            Seed to set internal randomness.

    Returns
    -------
        run: `whynot.dynamics.Run`
            Rollout of the model.

    """
    rng = np.random.RandomState(seed)

    # Iterate the discrete dynamics
    times = [config.start_time]
    states = [initial_state]
    state = copy.deepcopy(initial_state)
    for step in range(config.start_time, config.end_time):
        next_state = dynamics(state.values(), step, config, intervention, rng)
        state = State(*next_state)
        states.append(state)
        times.append(step + 1)

    return wn.dynamics.Run(states=states, times=times)


if __name__ == "__main__":
    print(simulate(State(), Config(end_time=20)))



"""Implementation of Epstein's Civil Violence model."""
import math

from mesa import Agent
from mesa import Model
from mesa.time import BaseScheduler
from mesa.space import Grid


class Citizen(Agent):
    # pylint: disable-msg=too-many-instance-attributes
    """A member of the general population, may or may not be in active rebellion.

    Summary of rule: If grievance - risk > threshold, rebel.

    Attributes
    ----------
        unique_id: unique int
        x, y: Grid coordinates
        hardship: Agent's 'perceived hardship (i.e., physical or economic
            privation).' Exogenous, drawn from U(0,1).
        regime_legitimacy: Agent's perception of regime legitimacy, equal
            across agents.  Exogenous.
        risk_aversion: Exogenous, drawn from U(0,1).
        threshold: if (grievance - (risk_aversion * arrest_probability)) >
            threshold, go/remain Active
        vision: number of cells in each direction (N, S, E and W) that agent
            can inspect
        condition: Can be "Quiescent" or "Active;" deterministic function of
            greivance, perceived risk, and
        grievance: deterministic function of hardship and regime_legitimacy;
            how aggrieved is agent at the regime?
        arrest_probability: agent's assessment of arrest probability, given
            rebellion

    """

    def __init__(
        self,
        unique_id,
        model,
        pos,
        hardship,
        regime_legitimacy,
        risk_aversion,
        threshold,
        vision,
    ):
        """Create a new Citizen.

        Parameters
        ----------
            unique_id: unique int
            x, y: Grid coordinates
            hardship: Agent's 'perceived hardship (i.e., physical or economic
                privation).' Exogenous, drawn from U(0,1).
            regime_legitimacy: Agent's perception of regime legitimacy, equal
                across agents.  Exogenous.
            risk_aversion: Exogenous, drawn from U(0,1).
            threshold: if (grievance - (risk_aversion * arrest_probability)) >
                threshold, go/remain Active
            vision: number of cells in each direction (N, S, E and W) that
                agent can inspect. Exogenous.
            model: model instance

        """
        super().__init__(unique_id, model)
        self.breed = "citizen"
        self.pos = pos
        self.hardship = hardship
        self.regime_legitimacy = regime_legitimacy
        self.risk_aversion = risk_aversion
        self.threshold = threshold
        self.condition = "Quiescent"
        self.vision = vision
        self.jail_sentence = 0
        self.grievance = lambda: self.hardship * (1 - self.regime_legitimacy)
        self.arrest_probability = None
        self.arrest_parameter = None
        self.arrests = 0
        self.prison_interaction = model.prison_interaction
        self.history = [self.grievance()]
        self.days_active = 0

        self.neighborhood = None
        self.neighbors = None
        self.empty_neighbors = None

    def update_arrest_parameter(self):
        """Change probability of arrest based on grievance."""
        self.arrest_parameter = (
            self.grievance() - self.arrest_probability * self.regime_legitimacy
        )

    def step(self):
        """Decide whether to activate, then move if applicable."""
        if self.jail_sentence:
            self.jail_sentence -= 1
            return  # no other changes or movements if agent is in jail.

        if self.condition == "Active":
            self.days_active += 1

        self.update_neighbors()
        self.update_estimated_arrest_probability()
        self.update_arrest_parameter()
        net_risk = self.risk_aversion * self.arrest_probability
        if (
            self.condition == "Quiescent"
            and (self.grievance() - net_risk) > self.threshold
        ):
            self.condition = "Active"
        elif (
            self.condition == "Active"
            and (self.grievance() - net_risk) <= self.threshold
        ):
            self.condition = "Quiescent"
        if self.model.movement and self.empty_neighbors:
            new_pos = self.random.choice(self.empty_neighbors)
            self.model.grid.move_agent(self, new_pos)

    def update_neighbors(self):
        """Look around and see who my neighbors are."""
        self.neighborhood = self.model.grid.get_neighborhood(
            self.pos, moore=True, radius=self.vision
        )
        self.neighbors = self.model.grid.get_cell_list_contents(self.neighborhood)
        self.empty_neighbors = [
            c for c in self.neighborhood if self.model.grid.is_cell_empty(c)
        ]

    def update_estimated_arrest_probability(self):
        """Estimate p(Arrest | I go active) based on local ratio of cops to actives."""
        cops_in_vision = len([c for c in self.neighbors if c.breed == "cop"])
        actives_in_vision = 1.0  # citizen counts herself
        for unit in self.neighbors:
            if (
                unit.breed == "citizen"
                and unit.condition == "Active"
                and unit.jail_sentence == 0
            ):
                actives_in_vision += 1
        self.arrest_probability = 1 - math.exp(
            -1 * self.model.arrest_prob_constant * (cops_in_vision / actives_in_vision)
        )

    def update_risk_aversion(self):
        """Change risk aversion for jailed citizens based on cellmates."""
        cellmates_risk_aversion = [
            c.risk_aversion
            for c in self.neighbors
            if ((c.breed == "citizen") and (c.jail_sentence > 0))
        ]
        if cellmates_risk_aversion:
            min_aversion = max(cellmates_risk_aversion)
            diff = self.prison_interaction * (min_aversion - self.risk_aversion)
            self.risk_aversion += diff


class Cop(Agent):
    # pylint: disable-msg=too-many-instance-attributes
    """A cop for life. No defection.

    Summary of rule: Inspect local vision and arrest a random active agent.

    Attributes
    ----------
        unique_id: unique int
        x, y: Grid coordinates
        vision: number of cells in each direction (N, S, E and W) that cop is
            able to inspect

    """

    def __init__(self, unique_id, model, pos, vision):
        """Create a new Cop.

        Parameters
        ----------
            unique_id: unique int
            x, y: Grid coordinates
            vision: number of cells in each direction (N, S, E and W) that
                agent can inspect. Exogenous.
            model: model instance

        """
        super().__init__(unique_id, model)
        self.breed = "cop"
        self.pos = pos
        self.vision = vision

        # Useless variables so BatchRunner wouldn't complain.
        self.jail_sentence = None
        self.condition = None
        self.arrest_probability = None
        self.arrests = None
        self.hardship = None
        self.regime_legitimacy = None
        self.risk_aversion = None
        self.threshold = None
        self.grievance = None
        self.arrests = None
        self.days_active = 0
        self.arrest_parameter = None

        self.neighborhood = None
        self.neighbors = None
        self.empty_neighbors = None

    def step(self):
        """Inspect neighborhood and arrest a random active agent. Move if applicable."""
        self.update_neighbors()
        active_neighbors = []
        for agent in self.neighbors:
            if (
                agent.breed == "citizen"
                and agent.condition == "Active"
                and agent.jail_sentence == 0
            ):
                active_neighbors.append(agent)
        if active_neighbors:
            arrestee = self.random.choice(active_neighbors)
            #             sentence = self.random.randint(0, self.model.max_jail_term)
            sentence = self.model.max_jail_term
            arrestee.jail_sentence = sentence
            arrestee.arrests += 1
            # arrestee is made quiescent if placed in Jail
            arrestee.condition = "Quiescent"
            # update perceived regime legitimacy if agent is arrested
            arrestee.update_risk_aversion()

        if self.model.movement and self.empty_neighbors:
            new_pos = self.random.choice(self.empty_neighbors)
            self.model.grid.move_agent(self, new_pos)

    def update_neighbors(self):
        """Look around and see who my neighbors are."""
        self.neighborhood = self.model.grid.get_neighborhood(
            self.pos, moore=True, radius=self.vision
        )
        self.neighbors = self.model.grid.get_cell_list_contents(self.neighborhood)
        self.empty_neighbors = [
            c for c in self.neighborhood if self.model.grid.is_cell_empty(c)
        ]


class RandomActivation(BaseScheduler):
    """A scheduler which activates each agent once per step, in random order.

    The order reshuffled every step.
    This is equivalent to the NetLogo 'ask agents...' and is generally the
    default behavior for an ABM. Assumes that all agents have a step(model) method.
    """

    def step(self):
        """Execute the step of all agents, one at a time, in random order."""
        for agent in self.agent_buffer(shuffled=True):
            agent.step()
        self.steps += 1
        self.time += 1


class CivilViolenceModel(Model):
    """Model encapsulates the entire civil violence simulator and interactions."""

    def __init__(
        self,
        height,
        width,
        cop_vision,
        max_jail_term,
        prison_interaction,
        arrest_prob_constant=2.3,
        movement=True,
        max_steps=1000,
        seed=None,
    ):
        """Seed is used to set randomness in the __new__ function of the Model superclass."""
        # pylint: disable-msg=unused-argument,super-init-not-called
        super().__init__()
        self.height = height
        self.width = width
        self.cop_vision = cop_vision
        self.max_jail_term = max_jail_term

        self.arrest_prob_constant = arrest_prob_constant
        self.movement = True
        self.running = True
        self.max_steps = max_steps
        self.iteration = 0
        self.prison_interaction = prison_interaction
        self.schedule = RandomActivation(self)
        self.grid = Grid(height, width, torus=True)

    def add_agent(
        self,
        unique_id,
        pos,
        hardship,
        legitimacy,
        risk_aversion,
        active_threshold,
        citizen_vision,
    ):
        """Add a new agent to the grid."""
        citizen = Citizen(
            unique_id,
            self,
            pos,
            hardship=hardship,
            regime_legitimacy=legitimacy,
            risk_aversion=risk_aversion,
            threshold=active_threshold,
            vision=citizen_vision,
        )
        x_coord, y_coord = pos
        self.grid[y_coord][x_coord] = citizen
        self.schedule.add(citizen)

    def add_cop(self, unique_id, pos):
        """Add a new copy to the grid."""
        cop = Cop(unique_id, self, pos, vision=self.cop_vision)
        x_coord, y_coord = pos
        self.grid[y_coord][x_coord] = cop
        self.schedule.add(cop)

    def find_empty(self):
        """Find a random empty location in the grid."""
        if self.grid.exists_empty_cells():
            pos = self.random.choice(sorted(self.grid.empties))
            return pos
        return None

    def step(self):
        """Advance the model by one step and collect data."""
        self.schedule.step()
        self.iteration += 1
        if self.iteration > self.max_steps:
            self.running = False



"""Gym module following OpenAI Gym style API for RL environments."""

import distutils.version
import os
import sys
import warnings

from gym import error
from gym.core import Env
from gym import logger

from whynot.gym.envs import make, spec, register

__all__ = ["Env", "make", "spec", "register"]



"""Tests for methods for working with dynamical systems."""
import copy
import dataclasses

import numpy as np
import pytest
import statsmodels

import whynot as wn
from whynot import parameter


def test_baseconfig():
    """Test the basic configuration object."""

    @dataclasses.dataclass
    class Config(wn.dynamics.BaseConfig):
        param1: float = 0
        param2: float = 23

    config = Config()

    assert config.parameter_names() == ["param1", "param2"]

    intervention = wn.dynamics.BaseIntervention(Config, 1970, param1=19)

    assert config.param1 == 0
    updated = config.update(intervention)
    assert config.param1 == 0
    assert config.param2 == 23
    assert updated.param1 == 19
    assert updated.param2 == 23


def test_basestate():
    """Test the basic state object."""

    @dataclasses.dataclass
    class State(wn.dynamics.BaseState):
        state1: float = 0
        state2: float = 1
        state3: float = 3

    assert State.num_variables() == 3
    assert State.variable_names() == ["state1", "state2", "state3"]

    state2 = [2]
    state = State(state2=state2)
    assert state.num_variables() == 3

    values = state.values()
    assert values[0] == 0
    assert values[1] == [2]
    assert values[2] == 3

    # Ensure values are shallow copied
    assert values[1] is state2


def test_framework_run():
    """Test the whynot.Run object"""
    states = ["start_state", 2, 3, 4, "final_state"]
    times = [4, 5, 6, 10, 23]

    run = wn.dynamics.Run(states=states, times=times)

    assert run.initial_state == "start_state"
    assert run.initial_time == 4
    assert run.final_state == "final_state"
    assert run.final_time == 23
    assert run[6] == 3

    # Find the state closest to time 11
    assert run[11] == 4

    # Throw error on construction if not the same length!
    with pytest.raises(ValueError):
        wn.dynamics.Run(states=[1, 2, 3], times=[0, 1])


@dataclasses.dataclass
class ToyConfig:
    num_steps: int = 10


@dataclasses.dataclass
class Intervention(ToyConfig):
    year: int = 0
    effect_size: int = 1


@dataclasses.dataclass
class ToyState:
    val: int = 0


class ToySimulator:
    """Simple simulator for testing purposes.

    Dynamics are to add effect_size to state at each time step if
    the run is treated. Otherwise, the run is constant starting at
    the initial_state value.

    """

    @staticmethod
    def simulate(initial_state, config, intervention=None, seed=None):
        """If treated, add a constant offset at each timestep."""
        states = []
        times = list(range(config.num_steps))

        state = initial_state
        for time in times:
            states.append(copy.deepcopy(state))
            if intervention and time >= intervention.year:
                state.val += intervention.effect_size
        return wn.dynamics.Run(states=states, times=times)


TEST_SIMULATOR = ToySimulator()


def check_shapes(dataset, num_samples, num_features):
    """Ensure the output has the right size."""
    assert dataset.covariates.shape == (num_samples, num_features)
    assert dataset.treatments.shape == (num_samples,)
    assert dataset.outcomes.shape == (num_samples,)
    assert dataset.true_effects.shape == (num_samples,)


def check_rct_assigment(assignment, prob):
    """Ensure that RCT assignment has approximately the correct proportion of treated."""
    total_treated = np.sum(assignment)
    ci_low, ci_upp = statsmodels.stats.proportion.proportion_confint(
        count=total_treated, nobs=len(assignment), alpha=0.001, method="beta"
    )
    assert ci_low <= prob <= ci_upp


### Helper methods for DynamicsExperiment tests
def basic_state_sampler():
    """Initial state is 0 for all rollouts."""
    return ToyState(val=0)


def correlated_state_sampler(num_samples):
    """Sample correlated initial states."""
    states = []
    for sample_idx in range(num_samples):
        states.append(ToyState(val=sample_idx))
    return states


def basic_outcome_extractor(run):
    """Look at state value after 5 steps."""
    return run[5].val


def basic_covariate_builder(run):
    """Return the final state of the run."""
    return np.array(run.states[-1].val)


def build_experiment(
    state_sampler=basic_state_sampler,
    config=ToyConfig(num_steps=10),
    intervention=Intervention(year=0, effect_size=1),
    propensity_scorer=0.5,
    outcome_extractor=basic_outcome_extractor,
    covariate_builder=basic_covariate_builder,
):
    """Construct a dynamics experiment instance with sane defaults."""
    return wn.dynamics.DynamicsExperiment(
        name="basic_test",
        description="Sanity check.",
        simulator=TEST_SIMULATOR,
        simulator_config=config,
        intervention=intervention,
        state_sampler=state_sampler,
        propensity_scorer=propensity_scorer,
        outcome_extractor=outcome_extractor,
        covariate_builder=covariate_builder,
    )


# DynamicsExperiment class tests.
def test_dynamics_experiment_basic():
    """Sanity check: does the class work?"""

    basic_exp = build_experiment()

    # Check outputs are the right sizes
    dataset = basic_exp.run(num_samples=10)
    check_shapes(dataset, 10, 1)

    dataset = basic_exp.run(num_samples=200)
    check_shapes(dataset, 200, 1)

    # Make sure the true effect is correctly computed
    assert np.allclose(5, dataset.true_effects)

    # Make sure the assignment and outcome are consistent
    assert np.allclose(0, dataset.outcomes[dataset.treatments == 0.0])
    assert np.allclose(5, dataset.outcomes[dataset.treatments == 1.0])

    # Make sure that incorrect parameters are flagged
    with pytest.raises(ValueError):
        basic_exp.run(num_samples=20, parameter=1234)


def test_state_sampler():
    """Test the state sampler function."""

    def initial_state_covariates(run):
        """Outcome is the initial state of the run."""
        return run.initial_state.val

    # Ensure the state sampler correctly uses randomness
    def random_initial_state(rng):
        """Initial state is 0 for all rollouts."""
        if rng.uniform() < 0.5:
            return ToyState(val=0)
        return ToyState(val=1)

    exp = build_experiment(
        state_sampler=random_initial_state, covariate_builder=initial_state_covariates
    )
    dataset0 = exp.run(num_samples=100, seed=1234)
    dataset1 = exp.run(num_samples=100, seed=1234)

    assert np.allclose(dataset0.covariates, dataset1.covariates)

    # Ensure the state sampler correctly deals with correlated state sampling
    exp = build_experiment(
        state_sampler=correlated_state_sampler,
        covariate_builder=initial_state_covariates,
    )
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(np.expand_dims(np.arange(100), axis=1), dataset.covariates)

    # Test that parameters play nicely with state sampler
    @parameter(name="init_val", default=2)
    def parameterized_sampler(init_val):
        return ToyState(val=init_val)

    exp = build_experiment(
        state_sampler=parameterized_sampler, covariate_builder=initial_state_covariates
    )

    # Ensure default works
    dataset = exp.run(num_samples=10)
    assert np.allclose(dataset.covariates, 2)

    # Ensure varying parameterization works
    dataset = exp.run(num_samples=10, init_val=3.14)
    assert np.allclose(dataset.covariates, 3.14)

    dataset = exp.run(num_samples=10, init_val=1234)
    assert np.allclose(dataset.covariates, 1234)

    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_val=123)


def test_config():
    """Test simulator config methods."""
    # Make sure config is actually effective
    def covariate_builder(run):
        return len(run.times)

    config = ToyConfig(num_steps=20)
    exp = build_experiment(config=config, covariate_builder=covariate_builder)
    dataset = exp.run(num_samples=15)
    assert np.allclose(20, dataset.covariates)

    # Check that config can be a function
    config = lambda: ToyConfig(num_steps=23)
    exp = build_experiment(config=config, covariate_builder=covariate_builder)
    dataset = exp.run(num_samples=15)
    assert np.allclose(23, dataset.covariates)

    # Check that these functions can be parameterized
    @parameter(name="num_steps", default=10)
    def pconfig(num_steps):
        return ToyConfig(num_steps=num_steps)

    exp = build_experiment(config=pconfig, covariate_builder=covariate_builder)

    # Default
    dataset = exp.run(num_samples=10)
    assert np.allclose(10, dataset.covariates)

    # Vary the parameter.
    dataset = exp.run(num_samples=10, num_steps=7)
    assert np.allclose(7, dataset.covariates)
    dataset = exp.run(num_samples=10, num_steps=56)
    assert np.allclose(56, dataset.covariates)

    # Check that bad argument flagging works for configs.
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_arg=123)


def test_intervention():
    """Test intervention methods."""
    intervention = Intervention(effect_size=2)
    exp = build_experiment(intervention=intervention)
    dataset = exp.run(num_samples=20)
    assert np.allclose(dataset.true_effects, 10)

    intervention = lambda: Intervention(year=1)
    exp = build_experiment(intervention=intervention)
    dataset = exp.run(num_samples=20)
    assert np.allclose(dataset.true_effects, 4)

    @parameter(name="effect_size", default=2)
    def pintervention(effect_size):
        return Intervention(effect_size=effect_size)

    exp = build_experiment(intervention=pintervention)

    # Default
    dataset = exp.run(num_samples=20)
    assert np.allclose(dataset.true_effects, 10)

    # Varying parameter
    dataset = exp.run(num_samples=20, effect_size=3)
    assert np.allclose(dataset.true_effects, 15)
    dataset = exp.run(num_samples=20, effect_size=5)
    assert np.allclose(dataset.true_effects, 25)

    # Check that bad argument flagging works for interventions
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_arg=123)


def test_propensity_scorer_rct():
    """Basic tests of propensity scorer functionality in RCT setting."""

    # Test to ensure propensity is (approximately) correct when using a constant
    exp = build_experiment(propensity_scorer=0.3)
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.3)

    exp = build_experiment(propensity_scorer=0.8)
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.8)

    # Make sure providing a function for the propensity score also works
    exp = build_experiment(propensity_scorer=lambda: 0.3)
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.3)

    # Only floats between 0 and 1 are valid constant propensities
    with pytest.raises(ValueError):
        exp = build_experiment(propensity_scorer="random_string")
        exp.run(num_samples=1000, seed=1234)

    with pytest.raises(ValueError):
        exp = build_experiment(propensity_scorer=2.0)
        exp.run(num_samples=1000, seed=1234)

    # Test parameterizing the propensity scorer
    @parameter(name="propensity", default=0.2)
    def parameterized_scorer(propensity):
        return propensity

    exp = build_experiment(propensity_scorer=parameterized_scorer)

    # Check default
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

    # Check using parameter
    dataset = exp.run(num_samples=1000, seed=1234, propensity=0.65)
    check_rct_assigment(dataset.treatments, 0.65)

    # Check error handing
    with pytest.raises(ValueError):
        exp.run(num_samples=10, seed=1234, random_arg=1234)

    # Tests specifying control_config and treatment_config
    config = ToyConfig(num_steps=13)

    def config_scorer(config):
        assert config.num_steps == 13
        return 0.2

    exp = build_experiment(propensity_scorer=config_scorer, config=config)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

    intervene = Intervention(year=4)

    def intervention_scorer(intervention):
        assert intervention.year == 4
        return 0.2

    exp = build_experiment(
        propensity_scorer=intervention_scorer, intervention=intervene
    )
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

    def both_scorer(config, intervention):
        assert config.num_steps == 13
        assert intervention.year == 4
        return 0.2

    exp = build_experiment(
        propensity_scorer=both_scorer, config=config, intervention=intervene
    )
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)


def test_propensity_scorer_observation():
    """Test propensity scorer in the observational setting."""

    # Test that specificying control_run, treatment_run gives you the correct
    # runs from the model.
    def crun_scorer(untreated_run):
        # Control runs are constant
        start = untreated_run.initial_state.val
        assert np.allclose(np.array([s.val for s in untreated_run.states]), start)
        return 0.5

    exp = build_experiment(
        state_sampler=correlated_state_sampler, propensity_scorer=crun_scorer
    )
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.5)

    def trun_scorer(treated_run):
        start = treated_run.initial_state.val
        run_vals = np.array([s.val for s in treated_run.states])
        assert np.allclose(start + np.arange(10), run_vals)
        return 0.5

    exp = build_experiment(
        state_sampler=correlated_state_sampler, propensity_scorer=trun_scorer
    )
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.5)

    # Check propensity scores with interference works
    def cruns_scorer(untreated_runs):
        # Make sure that we see all of the runs
        # control runs should start with 1, 2, 3, 4, ..., num_samples
        start_vals = sorted([run.initial_state.val for run in untreated_runs])
        assert np.allclose(np.arange(len(untreated_runs)), start_vals)

        # Make sure they are control runs (i.e. all constant)
        for run in untreated_runs:
            assert np.allclose(
                np.array([s.val for s in run.states]), run.initial_state.val
            )

        # Only treat first half of runs
        n = len(untreated_runs)
        propensities = np.zeros((n,))
        propensities[: n // 2] = 1.0
        return propensities

    n = 100
    exp = build_experiment(
        state_sampler=correlated_state_sampler, propensity_scorer=cruns_scorer
    )
    dataset = exp.run(num_samples=n, seed=1234)
    assert np.allclose(dataset.treatments[: n // 2], 1)
    assert np.allclose(dataset.treatments[n // 2 :], 0)

    def truns_scorer(treated_runs):
        # Make sure that we see all of the runs
        # control runs should start with 1, 2, 3, 4, ..., num_samples
        start_vals = sorted([run.initial_state.val for run in treated_runs])
        assert np.allclose(np.arange(len(treated_runs)), start_vals)

        # Make sure they are control runs (i.e. all constant)
        for run in treated_runs:
            assert np.allclose(
                np.array([s.val for s in run.states]),
                run.initial_state.val + np.arange(10),
            )

        # Only treat first half of runs
        n = len(treated_runs)
        propensities = np.zeros((n,))
        propensities[: n // 2] = 1.0
        return propensities

    exp = build_experiment(
        state_sampler=correlated_state_sampler, propensity_scorer=truns_scorer
    )
    dataset = exp.run(num_samples=n, seed=1234)
    assert np.allclose(dataset.treatments[: n // 2], 1)
    assert np.allclose(dataset.treatments[n // 2 :], 0)


def test_outcome_extractor():
    """Test the outcome extractor."""

    def compute_outcome(run):
        """Look at state value after 5 steps."""
        return run[5].val

    # Make sure compute outcome works as expected
    exp = build_experiment(outcome_extractor=compute_outcome)
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(dataset.outcomes[dataset.treatments == 0], 0)
    assert np.allclose(dataset.outcomes[dataset.treatments == 1], 5)

    # Compose compute_outcome with parameters
    @parameter(name="observation_step", default=3)
    def parameter_compute_outcome(run, observation_step):
        """Look at state value after observation_step steps."""
        return run[observation_step].val

    exp = build_experiment(outcome_extractor=parameter_compute_outcome)

    # Check defaults
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(dataset.outcomes[dataset.treatments == 0], 0)
    assert np.allclose(dataset.outcomes[dataset.treatments == 1], 3)

    # Check parameterized version
    dataset = exp.run(num_samples=100, seed=1234, observation_step=8)
    assert np.allclose(dataset.outcomes[dataset.treatments == 0], 0)
    assert np.allclose(dataset.outcomes[dataset.treatments == 1], 8)

    # Make sure error checking works
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_param=123)


def test_covariate_builder():
    """Simple tests for covariate builder function."""

    @parameter(name="step", default=4)
    def build_covariates(config, intervention, run, step):
        """Return the final state of the run."""
        # Make sure we're getting the correct config/intervention
        assert config.num_steps == 10 and intervention.effect_size == 1
        return run.states[step].val

    exp = build_experiment(covariate_builder=build_covariates)

    # Check defaults
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(dataset.covariates[dataset.treatments == 0], 0)
    assert np.allclose(dataset.covariates[dataset.treatments == 1], 4)

    # Check parameter
    dataset = exp.run(num_samples=100, seed=1234, step=7)
    assert np.allclose(dataset.covariates[dataset.treatments == 0], 0)
    assert np.allclose(dataset.covariates[dataset.treatments == 1], 7)

    # Make sure error checking works
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_param=123)



"""Implementation of Chen et al. systems dynamic model for opioid abuse.

The model implementation and parameterization is taken from:
    Chen Q, Larochelle MR, Weaver DT, et al. Prevention of Prescription Opioid
    Misuse and Projected Overdose Deaths in the United States. JAMA Network
    Open. 2019;2(2):e187621. doi:10.1001/jamanetworkopen.2018.7621
"""
import copy
import dataclasses

from scipy.integrate import odeint

import whynot as wn
import whynot.traceable_numpy as np
from whynot.dynamics import BaseConfig, BaseIntervention, BaseState


class TimeVaryingParam:  # pylint: disable=too-few-public-methods
    """Encapsulate a time-varying parameter obtained from a joinpoint model."""

    def __init__(
        self,
        baseline,
        alpha_1,
        joinpoint=None,
        alpha_2=None,
        intervention_year=None,
        intervention_val=0.0,
        start_year=2002,
        stabilize_year=None,
    ):
        """Specify a joinpoint model for a time-varying parameter.

        Parameters
        ----------
            baseline: float
                Baseline value at initial time point
            alpha_1: float
                Annual percentage change (APC) before joinpoint
            joinpoint: float
                (Optional) time indicating signifcant change in APC
            alpha_2: float
                (Optional) Annual percentage change (APC) after joinpoint
            intervention_year: float
                (Optional) Year to intervene on the APC
            intervention_val: float
                (Optional) New value for the APC after intevention. Default 0 is
                no change.
            start_year: float
                (Optional) initial time point
            stabilize_year: float
                (Optional) Year (after 2015) to set APC decrease to 0.

        """
        if joinpoint is None and alpha_2 is not None:
            raise ValueError("Must specify joinpoint to use alpha_2")

        if stabilize_year is not None and stabilize_year <= 2015:
            raise ValueError("stabilize_year must be after 2015")

        if intervention_year is not None:
            if stabilize_year is not None:
                raise ValueError(
                    "Cannot use stabilization and intervention simultaneously"
                )
            if joinpoint is not None and intervention_year < joinpoint:
                raise ValueError("Cannot intervene before joinpoint")

        self.baseline = baseline
        self.start_year = start_year
        self.alpha_1 = alpha_1
        self.joinpoint = joinpoint
        self.alpha_2 = alpha_2
        self.stabilize_year = stabilize_year
        self.intervention_year = intervention_year
        self.intervention_val = intervention_val

    def get_boxable(self):
        """Return the underlying parameter that can be boxed for dependency tracing."""
        return self.baseline

    def set_boxable(self, box):
        """Set the underlying parameter as a box for dependency tracing."""
        self.baseline = box

    def _unstabilized_value(self, time):
        """Return value of parameter without taking stabilization into account."""
        # pylint:disable-msg=no-member
        running_value = self.baseline
        if self.joinpoint is None or time <= self.joinpoint:
            return running_value * np.power(1.0 + self.alpha_1, time - self.start_year)

        running_value *= np.power(1.0 + self.alpha_1, self.joinpoint - self.start_year)

        if self.intervention_year is None or time <= self.intervention_year:
            return running_value * np.power(1.0 + self.alpha_2, time - self.joinpoint)

        running_value *= np.power(
            1.0 + self.alpha_2, self.intervention_year - self.joinpoint
        )

        return running_value * np.power(
            1.0 + self.intervention_val, time - self.intervention_year
        )

    def _current_rate(self, time):
        """Return current APC."""
        if self.joinpoint is None or time <= self.joinpoint:
            return self.alpha_1
        if self.intervention_year is None or time <= self.intervention_year:
            return self.alpha_2
        return self.intervention_val

    def __getitem__(self, time):
        """Retrieve the parameter value at time t."""
        if self.stabilize_year is None or time < 2015:
            return self._unstabilized_value(time)

        running_val = self._unstabilized_value(2015)
        rate = self._current_rate(time)

        # Decrease at a constant rate until stabilize year
        decay_per_year = rate / (self.stabilize_year - 2015)
        for year in range(2016, self.stabilize_year + 1):
            rate -= decay_per_year
            if time <= year:
                break
            running_val *= 1.0 + rate

        # Handle fractional values
        if time <= self.stabilize_year:
            # pylint:disable-msg=no-member
            remainder = time - np.floor(time)
            running_val *= (1.0 + rate) ** remainder

        return running_val


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of the dynamics for opioid epidemic simulator.

    By default, the confiruation uses the mean parameter values specified in
    eTable 4 and the pessimistic-case scenario (overdose mortality and incidence
    for illicit use stabilizes in 2025) and considers the scenario with no
    improvement in incidence after 2015.

    Examples
    --------
    >>> # Standard run of the dynamics with slightly rate of overdose
    >>> # deaths from nonmedial opioids.
    >>> opioid.Config(start_time=2002, end_time=2030, nonmedical_overdose=0.002)

    """

    # Dynamics parameters
    #: Annual incidence of nonmedical prescription opioid use.
    nonmedical_incidence: TimeVaryingParam = TimeVaryingParam(
        baseline=2496835, alpha_1=0.0016, joinpoint=2008, alpha_2=-0.0747
    )

    #: Annual incidence of illicit opioid use from sources other than prescription opioids.
    illicit_incidence: TimeVaryingParam = TimeVaryingParam(
        baseline=13489, alpha_1=0.235, stabilize_year=2025
    )

    #: Annual transition rate of non-medical opioid use to opioid use disorder.
    nonmedical_to_oud: float = 0.06

    #: Annual transition rate of users with opioid user disorder to illicit opioid users.
    oud_to_illicit: TimeVaryingParam = TimeVaryingParam(baseline=0.071, alpha_1=0.042)

    #: Annual transition rate of nonmedial to illicit opioid users
    nonmedical_to_illicit: float = 0.002

    #: Overdose mortality rate for nonmedial prescription opioid use.
    nonmedical_overdose: float = 0.001

    # Decrease baseline slightly from paper to match the reported
    # calibration targets.
    #: Overdose mortality rate for opioid use disorder.
    oud_overdose: TimeVaryingParam = TimeVaryingParam(
        baseline=0.003, alpha_1=0.224, joinpoint=2007, alpha_2=0.03
    )
    #: Overdose mortality rate for illicit opioid use.
    illicit_overdose: TimeVaryingParam = TimeVaryingParam(
        baseline=0.005,
        alpha_1=0.011,
        joinpoint=2011,
        alpha_2=0.356,
        stabilize_year=2025,
    )
    #: "Exit rate" from nonmedical users group. (Either stop using or die from non-opioid causes).
    nonmedical_exit: float = 0.177
    #: "Exit rate" from opioid user disorder group.
    oud_exit: TimeVaryingParam = TimeVaryingParam(baseline=0.311, alpha_1=-0.021)
    #: "Exit rate" of illicit opioid users.
    illicit_exit: float = 0.319

    # Simulator configuration
    #: Time to start the simulation (in years)
    start_time: float = 2002
    #: Time to end the simulation (in years)
    end_time: float = 2030
    #: Resolution of the recorded simulator output (in years)
    delta_t: float = 0.1

    def update(self, intervention):
        """Return a new config after applying intervention."""
        config = copy.deepcopy(self)
        if "nonmedical_incidence" in intervention.updates:
            nonmedical_incidence = intervention.updates["nonmedical_incidence"]
            config.nonmedical_incidence.intervention_year = intervention.time
            config.nonmedical_incidence.intervention_val = nonmedical_incidence
        if "illicit_incidence" in intervention.updates:
            illicit_incidence = intervention.updates["illicit_incidence"]
            config.illicit_incidence.intervention_year = intervention.time
            config.illicit_incidence.intervention_val = illicit_incidence
        return config


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Intervention parameters for opioid epidemic simulator.

    Currently, the simulator only supports interventions to reduce
    nonmedical and illicit opioid use because these are the only interventions
    considered in the originl Chen et al. paper.

    Examples
    --------
    >>> # Intervene in 2017 to reduce nonmedical incidence by 5%
    >>> opioid.Intervention(nonmedical_incidence=-0.05, time=2017)

    """

    def __init__(self, time=2015, nonmedical_incidence=None, illicit_incidence=None):
        """Construct an intervention object.

        Parameters
        ----------
            time: float
                time (in years) to intervene in the simulator.
            nonmedical_incidence: float
                Percent change in nonmedial opioid use (range -1. to 1.)
            illicit_incidence: float
                Percent change in illicit opioid use (range -1. to 1.)

        """
        kwargs = {}
        if nonmedical_incidence is not None:
            kwargs["nonmedical_incidence"] = nonmedical_incidence
        if illicit_incidence is not None:
            kwargs["illicit_incidence"] = illicit_incidence

        super(Intervention, self).__init__(Config, time, **kwargs)


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """State of the opioid simulator.

    Default values correspond to the mean initial state used in the Chen paper.
    """

    #: Non-medical users of prescription opioids
    nonmedical_users: float = 10029859
    #: Non-medical users of prescription opioid with prescription opioid use disorder.
    oud_users: float = 1369218
    #: Illicit opioid users (e.g. users of heroin/fentanyl as opposed to presciption opioids)
    illicit_users: float = 328371


def compute_overdose_deaths(run, start_year, end_year, config, intervention):
    """Compute number of opioid overdose deaths from start_year to end_year.

    Parameters
    ----------
        run: whynot.dynamics.Run
            Run object produced by running simulate for the opioid simulator
        start_year: int
            Year to start counting overdose deaths
        end_year: int
            Final year included in overdose death count
        config: whynot.simulator.opioid.Config
            Configuration used to generate `run`
        intervention: whynot.simulator.opioid.Intervention
            Intervention object to generate `run` when calling `simulate`

    Returns
    -------
        deaths: int
            Total number of overdose deaths across nonmedical users, users with
            OUD, and illicit opioid users.

    """
    if intervention:
        config = config.update(intervention)

    deaths = 0.0
    for outcome_year in range(start_year, end_year + 1):
        state = run[outcome_year]
        nonmedical_deaths = state.nonmedical_users * config.nonmedical_overdose
        oud_deaths = state.oud_users * config.oud_overdose[outcome_year]
        illicit_deaths = state.illicit_users * config.illicit_overdose[outcome_year]
        deaths += nonmedical_deaths + oud_deaths + illicit_deaths
    return deaths


def compute_illicit_deaths(run, outcome_year, config, intervention):
    """Compute number of overdose deaths for illicit use in year.

    Parameters
    ----------
        run: whynot.dynamics.Run
            Run object produced by running simulate for the opioid simulator
        outcome_year: int
            Year to measure illicit opioid use overdose deaths
        config: whynot.simulator.opioid.Config
            Configuration used to generate `run` when calling `simulate`
        intervention: whynot.simulator.opioid.Intervention
            Intervention object to generate `run` when calling `simulate`

    Returns
    -------
        deaths: int
            Number of overdose deaths for illicit opioid use in outcome_year.

    """
    if intervention:
        config = config.update(intervention)

    return run[outcome_year].illicit_users * config.illicit_overdose[outcome_year]


def dynamics(state, time, config, intervention=None):
    """State dynamics for the Chen et al. opioid epidemic model.

    Parameters
    ----------
        state: np.ndarray or tuple or list
            Instantenous system state
        time: float
            Time to evaluate the derivative
        config: whynot.simulators.opioid.Config
            Configuration parameters for the dynamics
        intervention: whynot.simulators.opioid.Intervention
            (Optional) Parameters specifying when/how to update the dynamics

    Returns
    -------
        ds_dt: np.ndarray
            Derivative of the state with respect to time, evaluated at state and time.

    """
    # In the opioid simulator, if there is an intervention,
    # it is applied for every time step of the dynamics (since
    # interventions are specified on TimeVaryingParams which internally track
    # the year).
    if intervention:
        config = config.update(intervention)

    nonmedical_users, oud_users, illicit_users = state

    # Update non-medical opioid use cases
    nonmedical_increase = config.nonmedical_incidence[time]
    nonmedical_decrease = nonmedical_users * (
        config.nonmedical_exit
        + config.nonmedical_overdose
        + config.nonmedical_to_oud
        + config.nonmedical_to_illicit
    )

    nonmedical_delta = nonmedical_increase - nonmedical_decrease

    # Update opioid use disorder cases
    oud_increase = config.nonmedical_to_oud * nonmedical_users
    oud_decrease = oud_users * (
        config.oud_exit[time] + config.oud_overdose[time] + config.oud_to_illicit[time]
    )
    oud_delta = oud_increase - oud_decrease

    # Update illicit opiate use cases
    illicit_increase = config.illicit_incidence[time] + (
        config.oud_to_illicit[time] * oud_users
        + config.nonmedical_to_illicit * nonmedical_users
    )

    illicit_decrease = illicit_users * (
        config.illicit_exit + config.illicit_overdose[time]
    )

    illicit_delta = illicit_increase - illicit_decrease

    return [nonmedical_delta, oud_delta, illicit_delta]


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Chen et al. model opioid epidemic model.

    Parameters
    ----------
        initial_state: whynot.simulator.opioid.State
            Initial state of the dynamics
        config: whynot.simulator.opioid.Config
            Config object to determine simulation dynamics.
        intervention: whynot.simulator.opioid.Intervention
            (Optional) Intervention object to determine what, if any,
            intervention to perform during the rollout of the dynamics.
        seed: int
            The simulator is deterministic, so the seed parameter is ignored.

    Returns
    -------
        run: whynot.dynamics.Run
            Run object produced by running simulate for the opioid simulator

    """
    # pylint: disable-msg=unused-argument
    # pylint:disable-msg=no-member
    t_eval = np.arange(
        config.start_time, config.end_time + config.delta_t, config.delta_t
    )

    solution = odeint(
        dynamics,
        y0=dataclasses.astuple(initial_state),
        t=t_eval,
        args=(config, intervention),
        rtol=1e-4,
        atol=1e-4,
    )

    states = [initial_state] + [State(*state) for state in solution[1:]]
    return wn.dynamics.Run(states=states, times=t_eval)



"""Simulation code for Lotka-Volterra model."""
import dataclasses

import numpy as np
from scipy.integrate import odeint

import whynot as wn
from whynot.dynamics import BaseConfig, BaseIntervention, BaseState


@dataclasses.dataclass
class Config(BaseConfig):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of Lotka-Volterra dynamics.

    Examples
    --------
    >>> # Configure the simulator so each caught rabbit creates 2 foxes
    >>> lotka_volterra.Config(fox_growth=0.5)

    """

    # Dynamics parameters
    #: Natural growth rate of rabbits, when there's no foxes.
    rabbit_growth: float = 1.0
    #: Natural death rate of rabbits, due to predation.
    rabbit_death: float = 0.1
    #: Natural death rate of fox, when there's no rabbits.
    fox_death: float = 1.5
    #: Factor describing how many caught rabbits create a new fox.
    fox_growth: float = 0.75

    # Simulator book-keeping
    #: Start time of the simulator (in years).
    start_time: float = 0
    #: End time of the simulator (in years).
    end_time: float = 100
    #: Spacing of the evaluation grid
    delta_t: float = 1.0


@dataclasses.dataclass
class State(BaseState):
    # pylint: disable-msg=too-few-public-methods
    """State of the Lotka-Volterra model."""

    #: Number of rabbits.
    rabbits: float = 10.0
    #: Number of foxes.
    foxes: float = 5.0


class Intervention(BaseIntervention):
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the Lotka-Volterra model.

    An intervention changes a subset of the configuration variables in the
    specified year. The remaining variables are unchanged.

    Examples
    --------
    >>> # Starting at time 25, set fox_growth to 0.4 (leave other variables unchanged)
    >>> Intervention(time=25, fox_growth=0.4)

    """

    def __init__(self, time=30, **kwargs):
        """Specify an intervention in lotka_volterra.

        Parameters
        ----------
            time: int
                Time of intervention in simulator dynamics.
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        super(Intervention, self).__init__(Config, time, **kwargs)


def dynamics(state, time, config, intervention=None):
    """Noisy Lotka-Volterra equations for 2 species.

    Equations from:
        https://scipy-cookbook.readthedocs.io/items/LoktaVolterraTutorial.html

    Parameters
    ----------
        state: np.ndarray or tuple or list
            Instantenous system state
        time: float
            Time to evaluate the derivative
        config: whynot.simulators.lotka_volterra.Config
            Configuration parameters for the dynamics
        intervention: whynot.simulators.lotka_volterra.Intervention
            (Optional) Parameters specifying when/how to update the dynamics

    Returns
    -------
        ds_dt: np.ndarray
            Derivative of the state with respect to time, evaluated at state and time.

    """
    if intervention and time >= intervention.time:
        config = config.update(intervention)

    rabbits, foxes = state

    delta_rabbits = (
        config.rabbit_growth * rabbits - config.rabbit_death * rabbits * foxes
    )

    delta_foxes = (
        -config.fox_death * foxes
        + config.fox_growth * config.rabbit_death * rabbits * foxes
    )

    ds_dt = np.array([delta_rabbits, delta_foxes])
    return ds_dt


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Lotka volterra model.

    Parameters
    ----------
        initial_state: whynot.lotka_volterra.State
        config: whynot.lotka_volterra.Config
            Base parameters for the simulator run
        intervention: whynot.lotka_volterra.Intervention
            (Optional) Parameters specifying a change in dynamics
        seed: int
            Unused since the simulator is deterministic.

    Returns
    -------
        run: whynot.dynamics.Run
            Simulator rollout

    """
    # pylint: disable-msg=unused-argument
    t_eval = np.arange(
        config.start_time, config.end_time + config.delta_t, config.delta_t
    )

    solution = odeint(
        dynamics,
        y0=dataclasses.astuple(initial_state),
        t=t_eval,
        args=(config, intervention),
        rtol=1e-4,
        atol=1e-4,
    )

    states = [initial_state] + [State(*state) for state in solution[1:]]
    return wn.dynamics.Run(states=states, times=t_eval)



"""Experiments for world3 simulator."""
import numpy as np

from whynot.dynamics import DynamicsExperiment
from whynot.framework import parameter
from whynot.simulators import world3

__all__ = [
    "get_experiments",
    "PollutionRCT",
    "PollutionConfounding",
    "PollutionUnobservedConfounding",
    "PollutionMediation",
]


def get_experiments():
    """Return all experiments for world3."""
    return [
        PollutionRCT,
        PollutionConfounding,
        PollutionUnobservedConfounding,
        PollutionMediation,
    ]


def sample_initial_states(rng):
    """Sample initial state by randomly perturbing the default state."""

    def random_scale(scale, base=10):
        """Make 10**(-scale) to 10**scale smaller/bigger uniformly."""
        return rng.choice(np.logspace(-scale, scale, base=base, num=50))

    state = world3.State()
    state.population_0_to_14 *= random_scale(0.33)
    state.population_15_to_44 *= random_scale(0.33)
    state.population_45_to_64 *= random_scale(0.33)
    state.population_65_and_over *= random_scale(0.33)
    state.industrial_capital *= random_scale(0.33)
    state.service_capital *= random_scale(0.33)
    state.arable_land *= random_scale(0.33)
    state.potentially_arable_land *= random_scale(0.33)
    state.urban_industrial_land *= random_scale(0.33)
    state.nonrenewable_resources *= random_scale(0.33)
    state.persistent_pollution *= random_scale(0.33)
    state.land_fertility *= random_scale(0.26)

    return state


##################
# RCT Experiment
##################
# pylint: disable-msg=invalid-name
#: An RCT experiment to study the effect of decreases in pollution generation on population.
PollutionRCT = DynamicsExperiment(
    name="PollutionRCT",
    description="Study effect of intervening in 1975 to decrease pollution generation.",
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.0),
    intervention=world3.Intervention(
        time=1975, persistent_pollution_generation_factor=0.85
    ),
    state_sampler=sample_initial_states,
    propensity_scorer=0.5,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=lambda run: run.initial_state.values(),
)


##########################
# Confounding Experiments
##########################
@parameter(
    name="treatment_bias",
    default=0.8,
    values=np.linspace(0.5, 1.0, 5),
    description="Treatment probability bias between low and high pollution runs.",
)
def pollution_confounded_propensity(intervention, untreated_runs, treatment_bias):
    """Probability of treating each unit.

    To generate confounding, we are more likely to treat worlds with high pollution.
    """

    def persistent_pollution(run):
        return run[intervention.time].persistent_pollution

    pollution = [persistent_pollution(run) for run in untreated_runs]
    upper_quantile = np.quantile(pollution, 0.9)

    def treatment_prob(idx):
        if pollution[idx] > upper_quantile:
            return treatment_bias
        return 1.0 - treatment_bias

    return np.array([treatment_prob(idx) for idx in range(len(untreated_runs))])


# pylint: disable-msg=invalid-name
#: An observational experiment with confounding. Polluted states are more likely to be treated.
PollutionConfounding = DynamicsExperiment(
    name="PollutionConfounding",
    description=(
        "Study effect of intervening to decrease pollution on total population. Confounding "
        "arises becauses states with high pollution are more likely "
        "to receive treatment."
    ),
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.0),
    intervention=world3.Intervention(
        time=1975, persistent_pollution_generation_factor=0.85
    ),
    state_sampler=sample_initial_states,
    propensity_scorer=pollution_confounded_propensity,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=lambda intervention, run: run[intervention.time].values(),
)


# pylint: disable-msg=invalid-name
#: An observational experiment with unobserved confounding.
PollutionUnobservedConfounding = DynamicsExperiment(
    name="PollutionUnobservedConfounding",
    description=(
        "Study effect of intervening to decrease pollution.  Confounding "
        "arises becauses states with high pollution are more likely "
        "to receive treatment. However, only variables that directly affect "
        "treatment assignment are observed."
    ),
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.0),
    intervention=world3.Intervention(
        time=1975, persistent_pollution_generation_factor=0.85
    ),
    state_sampler=sample_initial_states,
    propensity_scorer=pollution_confounded_propensity,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=lambda intervention, run: np.array(
        [run[intervention.time].persistent_pollution]
    ),
)


########################
# Mediation Experiments
########################
@parameter(
    name="mediation_year",
    default=2015,
    values=[1980, 2000, 2020, 2045],
    description="What year to get mediating variables",
)
@parameter(
    name="num_mediators",
    default=11,
    values=[0, 3, 5, 7, 10],
    description="number of mediators to include.",
)
def mediation_covariates(intervention, run, mediation_year, num_mediators):
    """Build the causal dataset."""
    # Just use all of the states at the moment of treatment assignment
    confounders = run[intervention.time].values()
    mediators = run[mediation_year].values()[:num_mediators]
    return np.concatenate([confounders, mediators])


# pylint: disable-msg=invalid-name
#: An observational experiment with mediation from states after intervention.
PollutionMediation = DynamicsExperiment(
    name="PollutionMediation",
    description=(
        "Study effect of intervening to decrease pollution.  Confounding "
        "arises becauses states with high pollution are more likely "
        "to receive treatment. Mediation arises since future states "
        "are also observed."
    ),
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.0),
    intervention=world3.Intervention(
        time=1975, persistent_pollution_generation_factor=0.85
    ),
    state_sampler=sample_initial_states,
    propensity_scorer=pollution_confounded_propensity,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=mediation_covariates,
)



"""World2 initialization."""

from whynot.simulators.world2.simulator import (
    Config,
    Intervention,
    quality_of_life,
    simulate,
    State,
)
from whynot.simulators.world2.experiments import *

SUPPORTS_CAUSAL_GRAPHS = False
